import argparse
import asyncio
import datetime
import importlib
import sys
from typing import Any, Dict, List, Optional, Type, cast

import structlog

import rasa.cli.utils
import rasa.shared.utils.cli
from rasa.cli import SubParsersAction
from rasa.cli.arguments.default_arguments import (
    add_endpoint_param,
    add_model_param,
    add_remote_storage_param,
)
from rasa.core.agent import Agent
from rasa.core.exceptions import AgentNotReady
from rasa.core.processor import MessageProcessor
from rasa.core.utils import AvailableEndpoints
from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding.generator import LLMBasedCommandGenerator
from rasa.dialogue_understanding.generator.command_parser import DEFAULT_COMMANDS
from rasa.dialogue_understanding_test.command_metric_calculation import (
    calculate_command_metrics,
)
from rasa.dialogue_understanding_test.constants import (
    DEFAULT_INPUT_TESTS_PATH,
    KEY_STUB_CUSTOM_ACTIONS,
)
from rasa.dialogue_understanding_test.du_test_result import (
    DialogueUnderstandingTestResult,
    DialogueUnderstandingTestSuiteResult,
)
from rasa.dialogue_understanding_test.du_test_runner import (
    DialogueUnderstandingTestRunner,
)
from rasa.dialogue_understanding_test.io import (
    read_test_suite,
    write_test_results_to_file,
)
from rasa.dialogue_understanding_test.validation import (
    validate_cli_arguments,
    validate_test_cases,
)
from rasa.e2e_test.e2e_test_case import TestSuite
from rasa.exceptions import RasaException
from rasa.shared.constants import (
    DEFAULT_ENDPOINTS_PATH,
    LLM_CONFIG_KEY,
    ROUTE_TO_CALM_SLOT,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.flows import FlowsList
from rasa.shared.utils.llm import (
    combine_custom_and_default_config,
    resolve_model_client_config,
)
from rasa.utils.beta import ensure_beta_feature_is_enabled
from rasa.utils.endpoints import EndpointConfig

RASA_PRO_BETA_DIALOGUE_UNDERSTANDING_TEST_ENV_VAR_NAME = (
    "RASA_PRO_BETA_DIALOGUE_UNDERSTANDING_TEST"
)

structlogger = structlog.get_logger()


def add_subparser(
    subparsers: SubParsersAction, parents: List[argparse.ArgumentParser]
) -> None:
    """Add the dialogue understanding test subparser to `rasa test`.

    Args:
        subparsers: subparser we are going to attach to
        parents: Parent parsers, needed to ensure tree structure in argparse
    """
    for subparser in subparsers.choices.values():
        if subparser.prog == "rasa test":
            du_test_subparser = create_du_test_subparser(parents)

            for action in subparser._subparsers._actions:
                if action.choices is not None:
                    action.choices["du"] = du_test_subparser
                    return

    # If we get here, we couldn't hook the subparser to `rasa test`
    raise RasaException(
        "Hooking the dialogue understanding (du) test subparser to "
        "`rasa test` command could not be completed. "
        "Cannot run dialogue understanding testing."
    )


def create_du_test_subparser(
    parents: List[argparse.ArgumentParser],
) -> argparse.ArgumentParser:
    """Create dialogue understanding test subparser."""
    du_test_subparser = argparse.ArgumentParser(
        prog="rasa test du",
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="Runs dialogue understanding testing.",
    )

    du_test_subparser.set_defaults(func=execute_dialogue_understanding_tests)

    add_du_test_arguments(du_test_subparser)
    add_bot_arguments(du_test_subparser)

    return du_test_subparser


def add_bot_arguments(parser: argparse.ArgumentParser) -> None:
    bot_arguments = parser.add_argument_group("Bot Settings")
    add_model_param(bot_arguments, add_positional_arg=False)
    add_endpoint_param(
        bot_arguments,
        help_text="Configuration file for the model server and the connectors as a "
        "yml file.",
    )
    add_remote_storage_param(bot_arguments)


def add_du_test_arguments(parser: argparse.ArgumentParser) -> None:
    """Arguments for running dialogue understanding tests."""
    du_arguments = parser.add_argument_group("Testing Settings")
    du_arguments.add_argument(
        "path-to-test-cases",
        nargs="?",
        type=str,
        default=DEFAULT_INPUT_TESTS_PATH,
        help="Input file or folder containing dialogue understanding test cases.",
    )
    du_arguments.add_argument(
        "--output-file",
        type=str,
        default="dialogue_understanding_test_{date:%Y%m%d-%H%M%S}.yml".format(
            date=datetime.datetime.now()
        ),
        help="Path to the output file to write the results to.",
    )
    du_arguments.add_argument(
        "--no-output",
        action="store_true",
        help="If set, no output file will be written to disk.",
    )
    du_arguments.add_argument(
        "--output-prompt",
        action="store_true",
        help="If set, the dialogue understanding test output will contain "
        "prompts for each failure.",
    )
    du_arguments.add_argument(
        "--additional-commands",
        type=str,
        nargs="*",
        help=(
            "List of additional custom command classes to add, separated by spaces. "
            "For example: --additional-commands my_module.MyCustomCommand"
        ),
    )
    du_arguments.add_argument(
        "--remove-default-commands",
        type=str,
        nargs="*",
        help=(
            f"List of default commands to remove, separated by spaces. "
            f"Default commands include: "
            f"{', '.join([command.__name__ for command in DEFAULT_COMMANDS])}. "
            f"For example: --remove-default-commands ClarifyCommand HumanHandoffCommand"
        ),
    )


def ensure_calm_only_bot(agent: Agent) -> None:
    if agent.domain is None or agent.processor is None:
        return

    if ROUTE_TO_CALM_SLOT in [slot.name for slot in agent.domain.slots]:
        rasa.shared.utils.cli.print_error(
            "You are using coexistence. Dialogue Understanding Tests do only work for "
            "CALM only assistants."
        )
        sys.exit(0)

    if not agent.processor.is_calm_assistant:
        rasa.shared.utils.cli.print_error(
            "Dialogue Understanding Tests do only work for CALM assistants. "
            "Your assistant does not use CALM."
        )
        sys.exit(0)


def execute_dialogue_understanding_tests(args: argparse.Namespace) -> None:
    """Run the dialogue understanding tests.

    Args:
        args: Commandline arguments.
    """
    ensure_beta_feature_is_enabled(
        "Dialogue Understanding (DU) Testing",
        env_flag=RASA_PRO_BETA_DIALOGUE_UNDERSTANDING_TEST_ENV_VAR_NAME,
    )

    # basic validation of the passed CLI arguments
    validate_cli_arguments(args)

    # initialization of endpoints
    endpoints = set_up_available_endpoints(args)

    # set up the test runner, e.g. start the agent
    try:
        test_runner = DialogueUnderstandingTestRunner(
            endpoints=endpoints,
            model_path=args.model,
            model_server=endpoints.model,
            remote_storage=args.remote_storage,
        )
    except AgentNotReady as error:
        structlogger.error(
            "rasa.dialogue_understanding_test.agent_not_ready", message=error.message
        )
        sys.exit(1)

    # Exit if the bot is not calm only
    ensure_calm_only_bot(test_runner.agent)

    # Ensure processor is not None so that we can extract the flows and the llm config
    if test_runner.agent.processor is None:
        rasa.shared.utils.cli.print_error(
            "No processor: Not able to retrieve flows and config from trained model."
        )
        sys.exit(0)

    # flows are needed in order to parse the commands when reading the test cases
    flows = asyncio.run(test_runner.agent.processor.get_flows())
    # llm config is needed for instrumentation
    llm_config = _get_llm_command_generator_config(test_runner.agent.processor)

    # read test cases from the given path
    test_suite: TestSuite = get_valid_test_suite(args, flows, test_runner.agent.domain)

    # setup stub custom actions if they are used
    set_up_stub_custom_actions(test_suite, endpoints)

    # run the actual test cases
    test_results = asyncio.run(
        test_runner.run_test_cases(
            test_suite.test_cases, test_suite.fixtures, test_suite.metadata
        )
    )

    # evaluate test results
    passing_test_results, failing_test_results = split_test_results(test_results)
    command_metrics = calculate_command_metrics(test_results)

    test_suite_result = DialogueUnderstandingTestSuiteResult.from_results(
        failing_test_results, passing_test_results, command_metrics, llm_config
    )

    # Do not move this import to the top of the file as it will break the
    # instrumentation of this function: the CLI module is initialized before the
    # instrumentation is set up, and we won't be able to "replace" the function
    # with the instrumented wrapper
    from rasa.dialogue_understanding_test.io import print_test_results

    # write results to console and file
    print_test_results(test_suite_result, output_prompt=args.output_prompt)
    if not args.no_output:
        write_test_results_to_file(
            test_suite_result,
            args.output_file,
            args.output_prompt,
        )


def _import_custom_command_class(class_path: str) -> Command:
    """Dynamically import a command class from a string path."""
    try:
        module_name, class_name = class_path.rsplit(".", 1)
        module = importlib.import_module(module_name)
        clz = getattr(module, class_name)
    except (ImportError, AttributeError, ValueError) as e:
        raise ValueError(f"Failed to import class '{class_path}': {e}")
    if not issubclass(clz, Command):
        structlogger.error(
            "rasa.dialogue_understanding_test.invalid_additional_command",
            event_info="The custom command class must be a subclass of Command.",
            class_path=class_path,
        )
        sys.exit(1)
    return clz


def _extract_additional_command_classes_from_cli_args(
    args: argparse.Namespace,
) -> List[Command]:
    """Extract additional command classes from the CLI arguments."""
    additional_commands = getattr(args, "additional_commands", [])
    if not additional_commands:
        return []
    return [
        _import_custom_command_class(command_module)
        for command_module in additional_commands
    ]


def get_valid_test_suite(
    args: argparse.Namespace, flows: FlowsList, domain: Optional[Domain]
) -> TestSuite:
    """Read the test cases from the given test case path and validate them."""
    path_to_test_cases = getattr(args, "path-to-test-cases", DEFAULT_INPUT_TESTS_PATH)
    remove_default_commands = getattr(args, "remove_default_commands", [])
    custom_command_classes = _extract_additional_command_classes_from_cli_args(args)
    test_suite = read_test_suite(
        path_to_test_cases, flows, custom_command_classes, remove_default_commands
    )
    validate_test_cases(test_suite.test_cases, domain)
    return test_suite


def set_up_available_endpoints(args: argparse.Namespace) -> AvailableEndpoints:
    """Set up the available endpoints for the test runner."""
    args.endpoints = rasa.cli.utils.get_validated_path(
        args.endpoints, "endpoints", DEFAULT_ENDPOINTS_PATH, True
    )
    endpoints = AvailableEndpoints.get_instance(args.endpoints)

    # Ignore all endpoints apart from action server, model, and nlu
    # to ensure InMemoryTrackerStore is being used instead of production
    # tracker store
    endpoints.tracker_store = None
    endpoints.lock_store = None
    endpoints.event_broker = None

    # disable nlg endpoint as we don't need it for dialogue understanding tests
    endpoints.nlg = None

    return endpoints


def set_up_stub_custom_actions(
    test_suite: TestSuite, endpoints: AvailableEndpoints
) -> None:
    """Set up the stub custom actions if they are used."""
    if test_suite.stub_custom_actions:
        if not endpoints.action:
            endpoints.action = EndpointConfig()

        endpoints.action.kwargs[KEY_STUB_CUSTOM_ACTIONS] = (
            test_suite.stub_custom_actions
        )


def split_test_results(
    results: List[DialogueUnderstandingTestResult],
) -> tuple[
    List[DialogueUnderstandingTestResult], List[DialogueUnderstandingTestResult]
]:
    """Split the test results into passed and failed test cases."""
    passed_cases = [r for r in results if r.passed]
    failed_cases = [r for r in results if not r.passed]

    return passed_cases, failed_cases


def _get_llm_command_generator_config(
    processor: MessageProcessor,
) -> Optional[Dict[str, Any]]:
    train_schema = processor.model_metadata.train_schema

    for node_name, node in train_schema.nodes.items():
        if node.matches_type(LLMBasedCommandGenerator, include_subtypes=True):
            # Configurations can reference model groups defined in the endpoints.yml
            resolved_llm_config = resolve_model_client_config(
                node.config.get(LLM_CONFIG_KEY, {}), node_name
            )
            llm_command_generator = cast(Type[LLMBasedCommandGenerator], node.uses)
            return combine_custom_and_default_config(
                resolved_llm_config, llm_command_generator.get_default_llm_config()
            )

    return None



import logging
import os
import typing
from typing import Dict, Text

import rasa.shared.utils.io
from rasa.constants import (
    ENV_CPU_INTER_OP_CONFIG,
    ENV_CPU_INTRA_OP_CONFIG,
    ENV_GPU_CONFIG,
)
from rasa.shared.utils import io as shared_io_utils
from rasa.utils.tensorflow.constants import TF_DETERMINISTIC_OPS

if typing.TYPE_CHECKING:
    from tensorflow import config as tf_config

logger = logging.getLogger(__name__)


def _setup_gpu_environment() -> None:
    """Sets configuration for TensorFlow GPU environment based on env variable."""
    gpu_memory_config = os.getenv(ENV_GPU_CONFIG)

    if not gpu_memory_config:
        return

    # Import from tensorflow only if necessary (environment variable was set)
    from tensorflow import config as tf_config

    parsed_gpu_config = _parse_gpu_config(gpu_memory_config)
    physical_gpus = tf_config.list_physical_devices("GPU")

    # Logic taken from https://www.tensorflow.org/guide/gpu
    if physical_gpus:
        for gpu_id, gpu_id_memory in parsed_gpu_config.items():
            _allocate_gpu_memory(physical_gpus[gpu_id], gpu_id_memory)

    else:
        rasa.shared.utils.io.raise_warning(
            f"You have an environment variable '{ENV_GPU_CONFIG}' set but no GPUs were "
            f"detected to configure."
        )


def _allocate_gpu_memory(
    gpu_instance: "tf_config.PhysicalDevice", logical_memory: int
) -> None:
    """Create a new logical device for the requested amount of memory.

    Args:
        gpu_instance: PhysicalDevice instance of a GPU device.
        logical_memory: Absolute amount of memory to be allocated to the new logical
            device.
    """
    from tensorflow import config as tf_config

    try:
        tf_config.experimental.set_virtual_device_configuration(
            gpu_instance,
            [
                tf_config.experimental.VirtualDeviceConfiguration(
                    memory_limit=logical_memory
                )
            ],
        )

    except RuntimeError:
        # Helper explanation of where the error comes from
        raise RuntimeError(
            "Error while setting up tensorflow environment. "
            "Virtual devices must be set before GPUs have been initialized."
        )


def _parse_gpu_config(gpu_memory_config: Text) -> Dict[int, int]:
    """Parse GPU configuration variable from a string to a dict.

    Args:
        gpu_memory_config: String containing the configuration for GPU usage.

    Returns:
        Parsed configuration as a dictionary with GPU IDs as keys and requested memory
        as the value.
    """
    # gpu_config is of format "gpu_id_1:gpu_id_1_memory, gpu_id_2: gpu_id_2_memory"
    # Parse it and store in a dictionary
    parsed_gpu_config: Dict[int, int] = {}

    try:
        for instance in gpu_memory_config.split(","):
            instance_gpu_id, instance_gpu_mem = instance.split(":")
            parsed_instance_gpu_id = int(instance_gpu_id)
            parsed_instance_gpu_mem = int(instance_gpu_mem)

            parsed_gpu_config[parsed_instance_gpu_id] = parsed_instance_gpu_mem
    except ValueError:
        # Helper explanation of where the error comes from
        raise ValueError(
            f"Error parsing GPU configuration. Please cross-check the format of "
            f"'{ENV_GPU_CONFIG}' at https://rasa.com/docs/rasa-pro/nlu-based-assistants/tuning-your-model"
            f"#restricting-absolute-gpu-memory-available ."
        )

    return parsed_gpu_config


def _setup_cpu_environment() -> None:
    """Set configuration for the CPU environment based on environment variable."""
    inter_op_parallel_threads = os.getenv(ENV_CPU_INTER_OP_CONFIG)
    intra_op_parallel_threads = os.getenv(ENV_CPU_INTRA_OP_CONFIG)

    if not inter_op_parallel_threads and not intra_op_parallel_threads:
        return

    from tensorflow import config as tf_config

    if inter_op_parallel_threads:
        try:
            inter_op_parallel_threads_number = int(inter_op_parallel_threads.strip())
        except ValueError:
            raise ValueError(
                f"Error parsing the environment variable '{ENV_CPU_INTER_OP_CONFIG}'. "
                f"Please cross-check the value."
            )

        tf_config.threading.set_inter_op_parallelism_threads(
            inter_op_parallel_threads_number
        )

    if intra_op_parallel_threads:
        try:
            intra_op_parallel_threads_number = int(intra_op_parallel_threads.strip())
        except ValueError:
            raise ValueError(
                f"Error parsing the environment variable '{ENV_CPU_INTRA_OP_CONFIG}'. "
                f"Please cross-check the value."
            )

        tf_config.threading.set_intra_op_parallelism_threads(
            intra_op_parallel_threads_number
        )


def setup_tf_environment() -> None:
    """Setup CPU and GPU related environment settings for TensorFlow."""
    _setup_cpu_environment()
    _setup_gpu_environment()


def check_deterministic_ops() -> None:
    """Warn user if they have set TF_DETERMINISTIC_OPS."""
    if os.getenv(TF_DETERMINISTIC_OPS, False):
        shared_io_utils.raise_warning(
            f"You have set '{TF_DETERMINISTIC_OPS}' to 1. If you are "
            f"using one or more GPU(s) and use any of 'SparseFeaturizer', "
            f"'TEDPolicy', 'DIETClassifier', 'UnexpecTEDIntentPolicy', or "
            f"'ResponseSelector' training and testing will fail as there are no "
            f"deterministic GPU implementations of some underlying TF ops.",
            category=UserWarning,
        )



from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import structlog

from rasa.dialogue_understanding.patterns.validate_slot import (
    ValidateSlotPatternFlowStackFrame,
)
from rasa.shared.constants import ACTION_ASK_PREFIX, UTTER_ASK_PREFIX
from rasa.shared.core.events import Event, SlotSet
from rasa.shared.core.slots import Slot
from rasa.shared.core.trackers import DialogueStateTracker

if TYPE_CHECKING:
    from rasa.dialogue_understanding.commands import StartFlowCommand
    from rasa.shared.core.flows import FlowsList

structlogger = structlog.get_logger()


def start_flow_by_name(
    flow_name: str, flows: "FlowsList"
) -> Optional["StartFlowCommand"]:
    from rasa.dialogue_understanding.commands import StartFlowCommand

    if flow_name in flows.user_flow_ids:
        return StartFlowCommand(flow=flow_name)
    else:
        structlogger.debug(
            "command_parser.start_flow_by_name.invalid_flow_id", flow=flow_name
        )
        return None


def extract_cleaned_options(options_str: str) -> List[str]:
    """Extract and clean options from a string."""
    delimiters = [",", " "]

    for delimiter in delimiters:
        options_str = options_str.replace(delimiter, " ")

    return sorted(
        opt.strip().strip('"').strip("'") for opt in options_str.split() if opt.strip()
    )


def is_none_value(value: str) -> bool:
    """Check if the value is a none value."""
    if not value:
        return True
    return value in {
        "[missing information]",
        "[missing]",
        "None",
        "undefined",
        "null",
    }


def clean_extracted_value(value: str) -> str:
    """Clean up the extracted value from the llm."""
    # replace any combination of single quotes, double quotes, and spaces
    # from the beginning and end of the string
    return value.strip("'\" ")


def get_nullable_slot_value(slot_value: str) -> Union[str, None]:
    """Get the slot value or None if the value is a none value.

    Args:
        slot_value: the value to coerce

    Returns:
        The slot value or None if the value is a none value.
    """
    return slot_value if not is_none_value(slot_value) else None


def initialize_pattern_validate_slot(
    slot: Slot,
) -> Optional[ValidateSlotPatternFlowStackFrame]:
    """Initialize the pattern to validate a slot value."""
    if not slot.requires_validation():
        return None

    validation = slot.validation
    slot_name = slot.name
    return ValidateSlotPatternFlowStackFrame(
        validate=slot_name,
        refill_utter=validation.refill_utter or f"{UTTER_ASK_PREFIX}{slot_name}",  # type: ignore[union-attr]
        refill_action=f"{ACTION_ASK_PREFIX}{slot_name}",
        rejections=validation.rejections,  # type: ignore[union-attr]
    )


def create_validate_frames_from_slot_set_events(
    tracker: DialogueStateTracker,
    events: List[Event],
    validate_frames: List[ValidateSlotPatternFlowStackFrame] = [],
    should_break: bool = False,
) -> Tuple[DialogueStateTracker, List[ValidateSlotPatternFlowStackFrame]]:
    """Process SlotSet events and create validation frames.

    Args:
        tracker: The dialogue state tracker.
        events: List of events to process.
        should_break:  whether or not to break after the first non-SlotSet event.
            if True, break out of the event loop as soon as the first non-SlotSet
            event is encountered.
            if False, continue processing the events until the end.

    Returns:
        Tuple of (updated tracker, list of validation frames).
    """
    for event in events:
        if not isinstance(event, SlotSet):
            if should_break:
                # we want to only process the most recent SlotSet events
                # so we break once we encounter a different event
                break
            continue

        slot = tracker.slots.get(event.key)
        frame = initialize_pattern_validate_slot(slot)

        if frame:
            validate_frames.append(frame)

    return tracker, validate_frames



import copy
import logging
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Text, cast

from jsonschema import Draft202012Validator

import rasa.core
import rasa.shared.utils.io
from rasa.core.actions.custom_action_executor import (
    CustomActionExecutor,
    NoEndpointCustomActionExecutor,
    RetryCustomActionExecutor,
)
from rasa.core.actions.direct_custom_actions_executor import DirectCustomActionExecutor
from rasa.core.actions.e2e_stub_custom_action_executor import (
    E2EStubCustomActionExecutor,
)
from rasa.core.actions.grpc_custom_action_executor import GRPCCustomActionExecutor
from rasa.core.actions.http_custom_action_executor import HTTPCustomActionExecutor
from rasa.core.constants import (
    KEY_IS_CALM_SYSTEM,
    KEY_IS_COEXISTENCE_ASSISTANT,
    UTTER_SOURCE_METADATA_KEY,
)
from rasa.core.policies.policy import PolicyPrediction
from rasa.core.utils import add_bot_utterance_metadata
from rasa.e2e_test.constants import KEY_STUB_CUSTOM_ACTIONS
from rasa.nlu.constants import (
    RESPONSE_SELECTOR_DEFAULT_INTENT,
    RESPONSE_SELECTOR_PREDICTION_KEY,
    RESPONSE_SELECTOR_PROPERTY_NAME,
    RESPONSE_SELECTOR_UTTER_ACTION_KEY,
)
from rasa.shared.constants import (
    ATTACHMENT,
    BUTTONS,
    CUSTOM,
    DEFAULT_NLU_FALLBACK_INTENT_NAME,
    DOCS_BASE_URL,
    ELEMENTS,
    FLOW_PREFIX,
    IMAGE,
    QUICK_REPLIES,
    ROUTE_TO_CALM_SLOT,
    TEXT,
    UTTER_PREFIX,
)
from rasa.shared.core.constants import (
    ACTION_BACK_NAME,
    ACTION_DEACTIVATE_LOOP_NAME,
    ACTION_DEFAULT_ASK_AFFIRMATION_NAME,
    ACTION_DEFAULT_ASK_REPHRASE_NAME,
    ACTION_DEFAULT_FALLBACK_NAME,
    ACTION_EXTRACT_SLOTS,
    ACTION_LISTEN_NAME,
    ACTION_METADATA_EXECUTION_ERROR_MESSAGE,
    ACTION_METADATA_EXECUTION_SUCCESS,
    ACTION_RESET_ROUTING,
    ACTION_RESTART_NAME,
    ACTION_REVERT_FALLBACK_EVENTS_NAME,
    ACTION_SEND_TEXT_NAME,
    ACTION_SESSION_START_NAME,
    ACTION_UNLIKELY_INTENT_NAME,
    DEFAULT_SLOT_NAMES,
    KNOWLEDGE_BASE_SLOT_NAMES,
    REQUESTED_SLOT,
    USER_INTENT_OUT_OF_SCOPE,
    SetSlotExtractor,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.events import (
    ActionExecuted,
    ActiveLoop,
    BotUttered,
    Event,
    Restarted,
    RoutingSessionEnded,
    SessionStarted,
    SlotSet,
    UserUtteranceReverted,
    UserUttered,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.slot_mappings import (
    SlotFillingManager,
    extract_slot_value,
)
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.exceptions import RasaException
from rasa.shared.nlu.constants import INTENT_NAME_KEY, INTENT_RANKING_KEY
from rasa.shared.utils.io import raise_warning
from rasa.shared.utils.schemas.events import EVENTS_SCHEMA
from rasa.utils.endpoints import EndpointConfig
from rasa.utils.url_tools import UrlSchema, get_url_schema

if TYPE_CHECKING:
    from rasa.core.channels.channel import OutputChannel
    from rasa.core.nlg import NaturalLanguageGenerator
    from rasa.shared.core.events import IntentPrediction

logger = logging.getLogger(__name__)


def default_actions(action_endpoint: Optional[EndpointConfig] = None) -> List["Action"]:
    """List default actions."""
    from rasa.core.actions.action_clean_stack import ActionCleanStack
    from rasa.core.actions.action_hangup import ActionHangup
    from rasa.core.actions.action_repeat_bot_messages import ActionRepeatBotMessages
    from rasa.core.actions.action_run_slot_rejections import ActionRunSlotRejections
    from rasa.core.actions.action_trigger_chitchat import ActionTriggerChitchat
    from rasa.core.actions.action_trigger_search import ActionTriggerSearch
    from rasa.core.actions.two_stage_fallback import TwoStageFallbackAction
    from rasa.dialogue_understanding.patterns.cancel import ActionCancelFlow
    from rasa.dialogue_understanding.patterns.clarify import ActionClarifyFlows
    from rasa.dialogue_understanding.patterns.correction import ActionCorrectFlowSlot

    return [
        ActionListen(),
        ActionRestart(),
        ActionSessionStart(),
        ActionDefaultFallback(),
        ActionDeactivateLoop(),
        ActionRevertFallbackEvents(),
        ActionDefaultAskAffirmation(),
        ActionDefaultAskRephrase(),
        TwoStageFallbackAction(action_endpoint),
        ActionUnlikelyIntent(),
        ActionSendText(),
        ActionBack(),
        ActionExtractSlots(action_endpoint),
        ActionCancelFlow(),
        ActionCorrectFlowSlot(),
        ActionClarifyFlows(),
        ActionRunSlotRejections(),
        ActionCleanStack(),
        ActionTriggerSearch(),
        ActionTriggerChitchat(),
        ActionResetRouting(),
        ActionHangup(),
        ActionRepeatBotMessages(),
    ]


def action_for_index(
    index: int,
    domain: Domain,
    action_endpoint: Optional[EndpointConfig],
) -> "Action":
    """Get an action based on its index in the list of available actions.

    Args:
        index: The index of the action. This is usually used by `Policy`s as they
            predict the action index instead of the name.
        domain: The `Domain` of the current model. The domain contains the actions
            provided by the user + the default actions.
        action_endpoint: Can be used to run `custom_actions`
            (e.g. using the `rasa-sdk`).

    Returns:
        The instantiated `Action` or `None` if no `Action` was found for the given
        index.
    """
    if domain.num_actions <= index or index < 0:
        raise IndexError(
            f"Cannot access action at index {index}. "
            f"Domain has {domain.num_actions} actions."
        )

    return action_for_name_or_text(
        domain.action_names_or_texts[index],
        domain,
        action_endpoint,
    )


def is_retrieval_action(action_name: Text, retrieval_intents: List[Text]) -> bool:
    """Check if an action name is a retrieval action.

    The name for a retrieval action has an extra `utter_` prefix added to
    the corresponding retrieval intent name.

    Args:
        action_name: Name of the action.
        retrieval_intents: List of retrieval intents defined in the NLU training data.

    Returns:
        `True` if the resolved intent name is present in the list of retrieval
        intents, `False` otherwise.
    """
    return (
        ActionRetrieveResponse.intent_name_from_action(action_name) in retrieval_intents
    )


def action_for_name_or_text(
    action_name_or_text: Text,
    domain: Domain,
    action_endpoint: Optional[EndpointConfig],
) -> "Action":
    """Retrieves an action by its name or by its text in case it's an end-to-end action.

    Args:
        action_name_or_text: The name of the action.
        domain: The current model domain.
        action_endpoint: The endpoint to execute custom actions.

    Raises:
        ActionNotFoundException: If action not in current domain.

    Returns:
        The instantiated action.
    """
    if action_name_or_text not in domain.action_names_or_texts:
        domain.raise_action_not_found_exception(action_name_or_text)

    defaults = {a.name(): a for a in default_actions(action_endpoint)}

    if (
        action_name_or_text in defaults
        and action_name_or_text not in domain.user_actions_and_forms
    ):
        return defaults[action_name_or_text]

    if action_name_or_text.startswith(UTTER_PREFIX) and is_retrieval_action(
        action_name_or_text, domain.retrieval_intents
    ):
        return ActionRetrieveResponse(action_name_or_text)

    if action_name_or_text in domain.action_texts:
        return ActionEndToEndResponse(action_name_or_text)

    if action_name_or_text.startswith(UTTER_PREFIX):
        return ActionBotResponse(action_name_or_text)

    is_form = action_name_or_text in domain.form_names
    # Users can override the form by defining an action with the same name as the form
    user_overrode_form_action = is_form and action_name_or_text in domain.user_actions
    if is_form and not user_overrode_form_action:
        from rasa.core.actions.forms import FormAction

        return FormAction(action_name_or_text, action_endpoint)

    if action_name_or_text.startswith(FLOW_PREFIX):
        from rasa.core.actions.action_trigger_flow import ActionTriggerFlow

        return ActionTriggerFlow(action_name_or_text)
    return RemoteAction(action_name_or_text, action_endpoint)


def create_bot_utterance(message: Dict[Text, Any]) -> BotUttered:
    """Create BotUttered event from message."""
    bot_message = BotUttered(
        text=message.pop(TEXT, None),
        data={
            ELEMENTS: message.pop(ELEMENTS, None),
            QUICK_REPLIES: message.pop(QUICK_REPLIES, None),
            BUTTONS: message.pop(BUTTONS, None),
            # for legacy / compatibility reasons we need to set the image
            # to be the attachment if there is no other attachment (the
            # `.get` is intentional - no `pop` as we still need the image`
            # property to set it in the following line)
            ATTACHMENT: message.pop(ATTACHMENT, None) or message.get(IMAGE, None),
            IMAGE: message.pop(IMAGE, None),
            CUSTOM: message.pop(CUSTOM, None),
        },
        metadata=message,
    )
    return bot_message


class Action:
    """Next action to be taken in response to a dialogue state."""

    def name(self) -> Text:
        """Unique identifier of this simple action."""
        raise NotImplementedError

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Execute the side effects of this action.

        Args:
            nlg: which ``nlg`` to use for response generation
            output_channel: ``output_channel`` to which to send the resulting message.
            tracker (DialogueStateTracker): the state tracker for the current
                user. You can access slot values using
                ``tracker.get_slot(slot_name)`` and the most recent user
                message is ``tracker.latest_message.text``.
            domain (Domain): the bot's domain
            metadata: Additional information for the action.

        Returns:
            A list of :class:`rasa.core.events.Event` instances
        """
        raise NotImplementedError

    def __str__(self) -> Text:
        """Returns text representation of form."""
        return f"{self.__class__.__name__}('{self.name()}')"

    def event_for_successful_execution(
        self,
        prediction: PolicyPrediction,
        was_successful: bool,
        error_message: Optional[str],
    ) -> ActionExecuted:
        """Event which should be logged for the successful execution of this action.

        Args:
            prediction: Prediction which led to the execution of this event.
            was_successful: Whether the action was executed successfully.
            error_message: Error message if the action was not executed successfully.

        Returns:
            Event which should be logged onto the tracker.
        """
        metadata = prediction.action_metadata or {}
        metadata[ACTION_METADATA_EXECUTION_SUCCESS] = was_successful
        metadata[ACTION_METADATA_EXECUTION_ERROR_MESSAGE] = error_message
        return ActionExecuted(
            self.name(),
            prediction.policy_name,
            prediction.max_confidence,
            hide_rule_turn=prediction.hide_rule_turn,
            metadata=metadata,
        )


class ActionBotResponse(Action):
    """An action which only effect is to utter a response when it is run."""

    def __init__(self, name: Text, silent_fail: Optional[bool] = False) -> None:
        """Creates action.

        Args:
            name: Name of the action.
            silent_fail: `True` if the action should fail silently in case no response
                was defined for this action.
        """
        self.utter_action = name
        self.silent_fail = silent_fail

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Simple run implementation uttering a (hopefully defined) response."""
        kwargs = {
            "domain_responses": domain.responses,
        }

        message = await nlg.generate(
            self.utter_action,
            tracker,
            output_channel.name(),
            **kwargs,
        )
        if message is None:
            if not self.silent_fail:
                logger.error(
                    "Couldn't create message for response '{}'.".format(
                        self.utter_action
                    )
                )
            return []

        message.update(metadata or {})
        message = add_bot_utterance_metadata(
            message, self.utter_action, nlg, domain, tracker
        )
        return [create_bot_utterance(message)]

    def name(self) -> Text:
        """Returns action name."""
        return self.utter_action


class ActionEndToEndResponse(Action):
    """Action to utter end-to-end responses to the user."""

    def __init__(self, action_text: Text) -> None:
        """Creates action.

        Args:
            action_text: Text of end-to-end bot response.
        """
        self.action_text = action_text

    def name(self) -> Text:
        """Returns action name."""
        # In case of an end-to-end action there is no label (aka name) for the action.
        # We fake a name by returning the text which the bot sends back to the user.
        return self.action_text

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action (see parent class for full docstring)."""
        message = {"text": self.action_text}
        return [create_bot_utterance(message)]

    def event_for_successful_execution(
        self,
        prediction: PolicyPrediction,
        was_successful: bool,
        error_message: Optional[str],
    ) -> ActionExecuted:
        """Event which should be logged for the successful execution of this action.

        Args:
            prediction: Prediction which led to the execution of this event.
            was_successful: Whether the action was executed successfully.
            error_message: Error message if the action was not executed successfully.

        Returns:
            Event which should be logged onto the tracker.
        """
        metadata = prediction.action_metadata or {}
        metadata[ACTION_METADATA_EXECUTION_SUCCESS] = was_successful
        metadata[ACTION_METADATA_EXECUTION_ERROR_MESSAGE] = error_message

        return ActionExecuted(
            policy=prediction.policy_name,
            confidence=prediction.max_confidence,
            action_text=self.action_text,
            hide_rule_turn=prediction.hide_rule_turn,
            metadata=metadata,
        )


class ActionRetrieveResponse(ActionBotResponse):
    """An action which queries the Response Selector for the appropriate response."""

    def __init__(self, name: Text, silent_fail: Optional[bool] = False) -> None:
        """Creates action. See docstring of parent class."""
        super().__init__(name, silent_fail)
        self.action_name = name
        self.silent_fail = silent_fail

    @staticmethod
    def intent_name_from_action(action_name: Text) -> Text:
        """Resolve the name of the intent from the action name."""
        return action_name.split(UTTER_PREFIX)[1]

    def get_full_retrieval_name(
        self, tracker: "DialogueStateTracker"
    ) -> Optional[Text]:
        """Returns full retrieval name for the action.

        Extracts retrieval intent from response selector and
        returns complete action utterance name.

        Args:
            tracker: Tracker containing past conversation events.

        Returns:
            Full retrieval name of the action if the last user utterance
            contains a response selector output, `None` otherwise.
        """
        latest_message = tracker.latest_message

        if latest_message is None:
            return None

        if RESPONSE_SELECTOR_PROPERTY_NAME not in latest_message.parse_data:
            return None

        response_selector_properties = latest_message.parse_data[
            RESPONSE_SELECTOR_PROPERTY_NAME  # type: ignore[literal-required]
        ]

        if (
            self.intent_name_from_action(self.action_name)
            in response_selector_properties
        ):
            query_key = self.intent_name_from_action(self.action_name)
        elif RESPONSE_SELECTOR_DEFAULT_INTENT in response_selector_properties:
            query_key = RESPONSE_SELECTOR_DEFAULT_INTENT
        else:
            return None

        selected = response_selector_properties[query_key]
        full_retrieval_utter_action = selected[RESPONSE_SELECTOR_PREDICTION_KEY][
            RESPONSE_SELECTOR_UTTER_ACTION_KEY
        ]
        return full_retrieval_utter_action

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Query the appropriate response and create a bot utterance with that."""
        latest_message = tracker.latest_message

        if latest_message is None:
            return []

        response_selector_properties = latest_message.parse_data[
            RESPONSE_SELECTOR_PROPERTY_NAME  # type: ignore[literal-required]
        ]

        if (
            self.intent_name_from_action(self.action_name)
            in response_selector_properties
        ):
            query_key = self.intent_name_from_action(self.action_name)
        elif RESPONSE_SELECTOR_DEFAULT_INTENT in response_selector_properties:
            query_key = RESPONSE_SELECTOR_DEFAULT_INTENT
        else:
            if not self.silent_fail:
                logger.error(
                    "Couldn't create message for response action '{}'.".format(
                        self.action_name
                    )
                )
            return []

        logger.debug(f"Picking response from selector of type {query_key}")
        selected = response_selector_properties[query_key]

        # Override utter action of ActionBotResponse
        # with the complete utter action retrieved from
        # the output of response selector.
        self.utter_action = selected[RESPONSE_SELECTOR_PREDICTION_KEY][
            RESPONSE_SELECTOR_UTTER_ACTION_KEY
        ]

        return await super().run(output_channel, nlg, tracker, domain)

    def name(self) -> Text:
        """Returns action name."""
        return self.action_name


class ActionBack(ActionBotResponse):
    """Revert the tracker state by two user utterances."""

    def name(self) -> Text:
        """Returns action back name."""
        return ACTION_BACK_NAME

    def __init__(self) -> None:
        """Initializes action back."""
        super().__init__("utter_back", silent_fail=True)

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        # only utter the response if it is available
        evts = await super().run(output_channel, nlg, tracker, domain)

        return evts + [UserUtteranceReverted(), UserUtteranceReverted()]


class ActionListen(Action):
    """The first action in any turn - bot waits for a user message.

    The bot should stop taking further actions and wait for the user to say
    something.
    """

    def name(self) -> Text:
        """Returns action listen name."""
        return ACTION_LISTEN_NAME

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        return []


class ActionResetRouting(Action):
    """Resets the tracker to its initial state.

    Utters the restart response if available.
    """

    def name(self) -> Text:
        """Returns action restart name."""
        return ACTION_RESET_ROUTING

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        # SlotSet(ROUTE_TO_CALM_SLOT, None) is needed to ensure the routing slot
        # is really reset to None and not just reset to an initial value
        return [RoutingSessionEnded(), SlotSet(ROUTE_TO_CALM_SLOT, None)]


class ActionRestart(ActionBotResponse):
    """Resets the tracker to its initial state.

    Utters the restart response if available.
    """

    def name(self) -> Text:
        """Returns action restart name."""
        return ACTION_RESTART_NAME

    def __init__(self) -> None:
        """Initializes action restart."""
        super().__init__("utter_restart", silent_fail=True)

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        # only utter the response if it is available
        evts = await super().run(output_channel, nlg, tracker, domain)

        return evts + [Restarted()]


class ActionSessionStart(Action):
    """Applies a conversation session start.

    Takes all `SlotSet` events from the previous session and applies them to the new
    session.
    """

    def name(self) -> Text:
        """Returns action start name."""
        return ACTION_SESSION_START_NAME

    @staticmethod
    def _slot_set_events_from_tracker(
        tracker: "DialogueStateTracker",
    ) -> List["SlotSet"]:
        """Fetch SlotSet events from tracker and carry over key, value and metadata."""
        return [
            SlotSet(key=event.key, value=event.value, metadata=event.metadata)
            for event in tracker.applied_events()
            if isinstance(event, SlotSet)
        ]

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        _events: List[Event] = [SessionStarted()]

        if domain.session_config.carry_over_slots:
            _events.extend(self._slot_set_events_from_tracker(tracker))

        _events.append(ActionExecuted(ACTION_LISTEN_NAME))

        return _events


class ActionDefaultFallback(ActionBotResponse):
    """Executes the fallback action and goes back to the prev state of the dialogue."""

    def name(self) -> Text:
        """Returns action default fallback name."""
        return ACTION_DEFAULT_FALLBACK_NAME

    def __init__(self) -> None:
        """Initializes action default fallback."""
        super().__init__("utter_default", silent_fail=True)

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        # only utter the response if it is available
        evts = await super().run(output_channel, nlg, tracker, domain)

        return evts + [UserUtteranceReverted()]


class ActionDeactivateLoop(Action):
    """Deactivates an active loop."""

    def name(self) -> Text:
        return ACTION_DEACTIVATE_LOOP_NAME

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        return [ActiveLoop(None), SlotSet(REQUESTED_SLOT, None)]


class RemoteActionJSONValidator:
    """Check action responses for JSON schema validation."""

    @staticmethod
    def action_response_format_spec() -> Dict[Text, Any]:
        """Expected response schema for an Action endpoint.

        Used for validation of the response returned from the
        Action endpoint.

        Returns:
            Dict[Text, Any]: A dictionary representing the JSON schema for validation.
        """
        schema = {
            "type": "object",
            "properties": {
                "events": EVENTS_SCHEMA,
                "responses": {"type": "array", "items": {"type": "object"}},
            },
        }
        return schema

    @classmethod
    def validate(cls, result: Dict[Text, Any]) -> bool:
        """Validate the given JSON result against the expected Action response schema.

        This method uses a cached JSON schema validator to check if the provided result
        conforms to the predefined schema.

        Args:
            result (Dict[Text, Any]): The JSON response to validate.

        Returns:
            bool: True if validation is successful.

        Raises:
            ValidationError: If the JSON response does not conform to the schema.
        """
        from jsonschema import ValidationError

        try:
            validator = cls.get_action_response_validator()
            validator.validate(
                result, RemoteActionJSONValidator.action_response_format_spec()
            )
            return True
        except ValidationError as e:
            e.message += (
                f". Failed to validate Action server response from API, "
                f"make sure your response from the Action endpoint is valid. "
                f"For more information about the format visit "
                f"{DOCS_BASE_URL}/action-server/actions"
            )
            raise e

    @classmethod
    @lru_cache(maxsize=1)
    def get_action_response_validator(cls) -> Draft202012Validator:
        """Retrieve a cached JSON schema validator for the Action response schema.

        Returns:
            Draft202012Validator: An instance of the JSON schema validator.
        """
        return Draft202012Validator(cls.action_response_format_spec())


class RemoteAction(Action):
    def __init__(
        self,
        name: Text,
        action_endpoint: Optional[EndpointConfig] = None,
    ) -> None:
        self._name = name
        self.action_endpoint = action_endpoint
        self.executor = self._create_executor()

    @lru_cache(maxsize=1)
    def _create_executor(self) -> CustomActionExecutor:
        """Creates an executor based on the action endpoint configuration.

        Returns:
            An instance of CustomActionExecutor.

        Raises:
            RasaException: If no valid action endpoint is configured.
        """
        if not self.action_endpoint:
            return NoEndpointCustomActionExecutor(self.name())

        if self.action_endpoint.kwargs.get(KEY_STUB_CUSTOM_ACTIONS):
            return E2EStubCustomActionExecutor(self.name(), self.action_endpoint)

        if self.action_endpoint.url and self.action_endpoint.actions_module:
            raise_warning(
                "Both 'actions_module' and 'url' are defined. "
                "As they are mutually exclusive and 'actions_module' "
                "is prioritized, actions will be executed by the assistant."
            )

        if self.action_endpoint and self.action_endpoint.actions_module:
            return DirectCustomActionExecutor(self.name(), self.action_endpoint)

        url_schema = get_url_schema(self.action_endpoint.url)

        if url_schema == UrlSchema.GRPC:
            return RetryCustomActionExecutor(
                GRPCCustomActionExecutor(self.name(), self.action_endpoint)
            )
        elif (
            url_schema == UrlSchema.HTTP
            or url_schema == UrlSchema.HTTPS
            or url_schema == UrlSchema.NOT_SPECIFIED
        ):
            return RetryCustomActionExecutor(
                HTTPCustomActionExecutor(self.name(), self.action_endpoint)
            )
        raise RasaException(
            f"Failed to create a custom action executor. "
            f"Please make sure to include an action endpoint configuration in your "
            f"endpoints configuration file. Make sure that for grpc, http and https "
            f"an url schema is set. "
            f"Found url '{self.action_endpoint.url}'."
        )

    @staticmethod
    async def _utter_responses(
        responses: List[Dict[Text, Any]],
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        **kwargs: Any,
    ) -> List[BotUttered]:
        """Use the responses generated by the action endpoint and utter them."""
        bot_messages = []
        domain: Domain = kwargs.get("domain", None)
        action_name: str = kwargs.get("action_name", None)
        for response in responses:
            generated_response = response.pop("response", None)
            if generated_response is not None:
                draft = await nlg.generate(
                    generated_response,
                    tracker,
                    output_channel.name(),
                    **response,
                )
                if not draft:
                    continue
                draft = add_bot_utterance_metadata(
                    draft, generated_response, nlg, domain, tracker
                )
            else:
                draft = {UTTER_SOURCE_METADATA_KEY: action_name}

            buttons = response.pop("buttons", []) or []
            if buttons:
                draft.setdefault("buttons", [])
                draft["buttons"].extend(buttons)

            # Avoid overwriting `draft` values with empty values
            response = {k: v for k, v in response.items() if v is not None}
            draft.update(response)
            bot_messages.append(create_bot_utterance(draft))

        return bot_messages

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        response = await self.executor.run(
            domain=domain,
            tracker=tracker,
        )

        events_json = response.get("events", [])
        responses = response.get("responses", [])
        bot_messages = await self._utter_responses(
            responses,
            output_channel,
            nlg,
            tracker,
            domain=domain,
            action_name=self.name(),
        )

        events = rasa.shared.core.events.deserialise_events(events_json)

        processed_events = []
        for event in events:
            if isinstance(event, SlotSet) and event.filled_by is None:
                event.filled_by = SetSlotExtractor.CUSTOM.value
            processed_events.append(event)

        return cast(List[Event], bot_messages) + processed_events

    def name(self) -> Text:
        return self._name


class ActionRevertFallbackEvents(Action):
    """Reverts events which were done during the `TwoStageFallbackPolicy`.

    This reverts user messages and bot utterances done during a fallback
    of the `TwoStageFallbackPolicy`. By doing so it is not necessary to
    write custom stories for the different paths, but only of the happy
    path. This is deprecated and can be removed once the
    `TwoStageFallbackPolicy` is removed.
    """

    def name(self) -> Text:
        return ACTION_REVERT_FALLBACK_EVENTS_NAME

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        from rasa.core.policies.two_stage_fallback import has_user_rephrased

        # User rephrased
        if has_user_rephrased(tracker):
            return _revert_successful_rephrasing(tracker)
        # User affirmed
        elif has_user_affirmed(tracker):
            return _revert_affirmation_events(tracker)
        else:
            return []


class ActionUnlikelyIntent(Action):
    """An action that indicates that the intent predicted by NLU is unexpected.

    This action can be predicted by `UnexpecTEDIntentPolicy`.
    """

    def name(self) -> Text:
        """Returns the name of the action."""
        return ACTION_UNLIKELY_INTENT_NAME

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        return []


def has_user_affirmed(tracker: "DialogueStateTracker") -> bool:
    """Indicates if the last executed action is `action_default_ask_affirmation`."""
    return tracker.last_executed_action_has(ACTION_DEFAULT_ASK_AFFIRMATION_NAME)


def _revert_affirmation_events(tracker: "DialogueStateTracker") -> List[Event]:
    revert_events = _revert_single_affirmation_events()

    # User affirms the rephrased intent
    rephrased_intent = tracker.last_executed_action_has(
        name=ACTION_DEFAULT_ASK_REPHRASE_NAME, skip=1
    )
    if rephrased_intent:
        revert_events += _revert_rephrasing_events()

    last_user_event = tracker.get_last_event_for(UserUttered)
    if not last_user_event:
        raise TypeError("Cannot find last event to revert to.")

    last_user_event = copy.deepcopy(last_user_event)
    # FIXME: better type annotation for `parse_data` would require
    # a larger refactoring (e.g. switch to dataclass)
    last_user_event.parse_data["intent"]["confidence"] = 1.0  # type: ignore[typeddict-item]

    return revert_events + [last_user_event]


def _revert_single_affirmation_events() -> List[Event]:
    return [
        UserUtteranceReverted(),  # revert affirmation and request
        # revert original intent (has to be re-added later)
        UserUtteranceReverted(),
        # add action listen intent
        ActionExecuted(action_name=ACTION_LISTEN_NAME),
    ]


def _revert_successful_rephrasing(tracker: "DialogueStateTracker") -> List[Event]:
    last_user_event = tracker.get_last_event_for(UserUttered)
    if not last_user_event:
        raise TypeError("Cannot find last event to revert to.")

    last_user_event = copy.deepcopy(last_user_event)
    return _revert_rephrasing_events() + [last_user_event]


def _revert_rephrasing_events() -> List[Event]:
    return [
        UserUtteranceReverted(),  # remove rephrasing
        # remove feedback and rephrase request
        UserUtteranceReverted(),
        # remove affirmation request and false intent
        UserUtteranceReverted(),
        # replace action with action listen
        ActionExecuted(action_name=ACTION_LISTEN_NAME),
    ]


# TODO: this should be removed, e.g. it uses a hardcoded message and no translation
class ActionDefaultAskAffirmation(Action):
    """Default implementation which asks the user to affirm his intent.

    It is suggested to overwrite this default action with a custom action
    to have more meaningful prompts for the affirmations. E.g. have a
    description of the intent instead of its identifier name.
    """

    def name(self) -> Text:
        return ACTION_DEFAULT_ASK_AFFIRMATION_NAME

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        latest_message = tracker.latest_message
        if latest_message is None:
            raise TypeError(
                "Cannot find last user message for detecting fallback affirmation."
            )

        intent_to_affirm = latest_message.intent.get(INTENT_NAME_KEY)

        # FIXME: better type annotation for `parse_data` would require
        # a larger refactoring (e.g. switch to dataclass)
        intent_ranking = cast(
            List["IntentPrediction"],
            latest_message.parse_data.get(INTENT_RANKING_KEY) or [],
        )
        if (
            intent_to_affirm == DEFAULT_NLU_FALLBACK_INTENT_NAME
            and len(intent_ranking) > 1
        ):
            intent_to_affirm = intent_ranking[1][INTENT_NAME_KEY]  # type: ignore[literal-required]

        affirmation_message = f"Did you mean '{intent_to_affirm}'?"

        message = {
            "text": affirmation_message,
            "buttons": [
                {"title": "Yes", "payload": f"/{intent_to_affirm}"},
                {"title": "No", "payload": f"/{USER_INTENT_OUT_OF_SCOPE}"},
            ],
            "utter_action": self.name(),
        }

        return [create_bot_utterance(message)]


class ActionDefaultAskRephrase(ActionBotResponse):
    """Default implementation which asks the user to rephrase his intent."""

    def name(self) -> Text:
        """Returns action default ask rephrase name."""
        return ACTION_DEFAULT_ASK_REPHRASE_NAME

    def __init__(self) -> None:
        """Initializes action default ask rephrase."""
        super().__init__("utter_ask_rephrase", silent_fail=True)


class ActionSendText(Action):
    """Sends a text message to the output channel."""

    def name(self) -> Text:
        return ACTION_SEND_TEXT_NAME

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        fallback = {"text": ""}
        metadata_copy = copy.deepcopy(metadata) if metadata else {}
        message = metadata_copy.get("message", fallback)
        return [create_bot_utterance(message)]


class ActionExtractSlots(Action):
    """Default action that runs after each user turn.

    Action is executed automatically in MessageProcessor.handle_message(...)
    before the next predicted action is run.

    Set slots to extracted values from user message
    according to assigned slot mappings.
    """

    def __init__(self, action_endpoint: Optional[EndpointConfig]) -> None:
        """Initializes default action extract slots."""
        self._action_endpoint = action_endpoint

    def name(self) -> Text:
        """Returns action_extract_slots name."""
        return ACTION_EXTRACT_SLOTS

    async def run(
        self,
        output_channel: "OutputChannel",
        nlg: "NaturalLanguageGenerator",
        tracker: "DialogueStateTracker",
        domain: "Domain",
        metadata: Optional[Dict[Text, Any]] = None,
    ) -> List[Event]:
        """Runs action. Please see parent class for the full docstring."""
        slot_events: List[Event] = []
        user_slots = [
            slot
            for slot in domain.slots
            if slot.name not in DEFAULT_SLOT_NAMES | KNOWLEDGE_BASE_SLOT_NAMES
        ]

        all_flows = metadata.get("all_flows") if metadata else None
        flows = FlowsList.from_json(all_flows) if all_flows else None
        calm_slot_names = flows.available_slot_names() if flows else set()
        is_calm_system = metadata.get(KEY_IS_CALM_SYSTEM) if metadata else False
        is_coexistence_bot = (
            metadata.get(KEY_IS_COEXISTENCE_ASSISTANT) if metadata else False
        )
        slot_filling_manager = SlotFillingManager(
            domain,
            tracker,
            action_endpoint=self._action_endpoint,
        )

        for slot in user_slots:
            # allows the action to set slots that are shared between
            # the NLU-based system and CALM system in a coexistence bot
            if is_coexistence_bot:
                should_fill_slot_in_coexistence = (
                    slot_filling_manager.should_fill_slot_in_coexistence(
                        is_calm_system=is_calm_system,
                        slot=slot,
                        calm_slot_names=calm_slot_names,
                    )
                )
                if not should_fill_slot_in_coexistence:
                    continue

            if not is_calm_system:
                slot_value, is_extracted = extract_slot_value(
                    slot, slot_filling_manager
                )
                if is_extracted:
                    slot_events.append(SlotSet(slot.name, slot_value))

            custom_events = await slot_filling_manager.run_action_at_every_turn(
                slot,
                output_channel,
                nlg,
            )
            slot_events.extend(custom_events)

        if not is_calm_system:
            validated_events = await slot_filling_manager.execute_validation_action(
                slot_events,
                output_channel,
                nlg,
            )
            return validated_events

        return slot_events



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from typing import Any, Dict, Optional

import structlog

from rasa.shared.constants import (
    API_BASE_CONFIG_KEY,
    API_TYPE_CONFIG_KEY,
    API_VERSION_CONFIG_KEY,
    LANGCHAIN_TYPE_CONFIG_KEY,
    MAX_COMPLETION_TOKENS_CONFIG_KEY,
    MAX_TOKENS_CONFIG_KEY,
    MODEL_CONFIG_KEY,
    MODEL_NAME_CONFIG_KEY,
    N_REPHRASES_CONFIG_KEY,
    OPENAI_API_BASE_CONFIG_KEY,
    OPENAI_API_TYPE,
    OPENAI_API_TYPE_CONFIG_KEY,
    OPENAI_API_VERSION_CONFIG_KEY,
    OPENAI_PROVIDER,
    PROVIDER_CONFIG_KEY,
    RASA_TYPE_CONFIG_KEY,
    REQUEST_TIMEOUT_CONFIG_KEY,
    STREAM_CONFIG_KEY,
    TIMEOUT_CONFIG_KEY,
)
from rasa.shared.providers._configs.utils import (
    raise_deprecation_warnings,
    resolve_aliases,
    validate_forbidden_keys,
    validate_required_keys,
)

structlogger = structlog.get_logger()


DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING = {
    # Model name aliases
    MODEL_NAME_CONFIG_KEY: MODEL_CONFIG_KEY,
    # Provider aliases
    RASA_TYPE_CONFIG_KEY: PROVIDER_CONFIG_KEY,
    LANGCHAIN_TYPE_CONFIG_KEY: PROVIDER_CONFIG_KEY,
    # API type aliases
    OPENAI_API_TYPE_CONFIG_KEY: API_TYPE_CONFIG_KEY,
    # API base aliases
    OPENAI_API_BASE_CONFIG_KEY: API_BASE_CONFIG_KEY,
    # API version aliases
    OPENAI_API_VERSION_CONFIG_KEY: API_VERSION_CONFIG_KEY,
    # Timeout aliases
    REQUEST_TIMEOUT_CONFIG_KEY: TIMEOUT_CONFIG_KEY,
    # Max tokens aliases
    MAX_TOKENS_CONFIG_KEY: MAX_COMPLETION_TOKENS_CONFIG_KEY,
}

REQUIRED_KEYS = [MODEL_CONFIG_KEY]

FORBIDDEN_KEYS = [
    STREAM_CONFIG_KEY,
    N_REPHRASES_CONFIG_KEY,
]


@dataclass
class OpenAIClientConfig:
    """Parses configuration for Azure OpenAI client, resolves aliases and
    raises deprecation warnings.

    Raises:
        ValueError: Raised in cases of invalid configuration:
            - If any of the required configuration keys are missing.
            - If `api_type` has a value different from `openai`.
    """

    model: str
    api_base: Optional[str]
    api_version: Optional[str]

    # API Type is not actually used by LiteLLM backend, but we define
    # it here for backward compatibility.
    api_type: str = OPENAI_API_TYPE

    # Provider is not used by LiteLLM backend, but we define
    # it here since it's used as switch between different
    # clients
    provider: str = OPENAI_PROVIDER

    extra_parameters: dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        # In case of OpenAI hosting, it doesn't make sense
        # for API type to be anything else that 'openai'
        if self.api_type != OPENAI_API_TYPE:
            message = f"API type must be set to '{OPENAI_API_TYPE}'."
            structlogger.error(
                "openai_client_config.validation_error",
                message=message,
                api_type=self.api_type,
            )
            raise ValueError(message)
        if self.provider != OPENAI_PROVIDER:
            message = f"Provider must be set to '{OPENAI_PROVIDER}'."
            structlogger.error(
                "openai_client_config.validation_error",
                message=message,
                provider=self.provider,
            )
            raise ValueError(message)
        if self.model is None:
            message = "Model cannot be set to None."
            structlogger.error(
                "openai_client_config.validation_error",
                message=message,
                model=self.model,
            )
            raise ValueError(message)

    @classmethod
    def from_dict(cls, config: dict) -> OpenAIClientConfig:
        """
        Initializes a dataclass from the passed config.

        Args:
            config: (dict) The config from which to initialize.

        Raises:
            ValueError: Config is missing required keys.

        Returns:
            AzureOpenAIClientConfig
        """
        # Check for deprecated keys
        raise_deprecation_warnings(config, DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING)
        # Resolve any potential aliases
        config = cls.resolve_config_aliases(config)
        # Validate that the required keys are present
        validate_required_keys(config, REQUIRED_KEYS)
        # Validate that the forbidden keys are not present
        validate_forbidden_keys(config, FORBIDDEN_KEYS)
        this = OpenAIClientConfig(
            # Required parameters
            model=config.pop(MODEL_CONFIG_KEY),
            # Pop the 'provider' key. Currently, it's *optional* because of
            # backward compatibility with older versions.
            provider=config.pop(PROVIDER_CONFIG_KEY, OPENAI_PROVIDER),
            # Optional parameters
            api_base=config.pop(API_BASE_CONFIG_KEY, None),
            api_version=config.pop(API_VERSION_CONFIG_KEY, None),
            api_type=config.pop(API_TYPE_CONFIG_KEY, OPENAI_API_TYPE),
            # The rest of parameters (e.g. model parameters) are considered
            # as extra parameters (this also includes timeout).
            extra_parameters=config,
        )
        return this

    def to_dict(self) -> dict:
        """Converts the config instance into a dictionary."""
        d = asdict(self)
        # Extra parameters should also be on the top level
        d.pop("extra_parameters", None)
        d.update(self.extra_parameters)
        return d

    @staticmethod
    def resolve_config_aliases(config: Dict[str, Any]) -> Dict[str, Any]:
        return resolve_aliases(config, DEPRECATED_ALIASES_TO_STANDARD_KEY_MAPPING)


def is_openai_config(config: dict) -> bool:
    """Check whether the configuration is meant to configure
    an OpenAI client.
    """
    # Process the config to handle all the aliases
    config = OpenAIClientConfig.resolve_config_aliases(config)

    # Case: Configuration contains `provider: openai`
    if config.get(PROVIDER_CONFIG_KEY) == OPENAI_PROVIDER:
        return True

    return False



import logging
from collections import OrderedDict
from typing import Optional

from rasa.core.channels.voice_stream.audio_bytes import RasaAudioBytes

logger = logging.getLogger(__name__)


class TTSCache:
    """An LRU Cache for TTS based on pythons OrderedDict."""

    def __init__(self, max_size: int):
        self.cache: OrderedDict[str, RasaAudioBytes] = OrderedDict()
        self.max_size = max_size

    def get(self, text: str) -> Optional[RasaAudioBytes]:
        if text not in self.cache:
            return None
        else:
            self.cache.move_to_end(text)
            return self.cache[text]

    def put(self, text: str, audio_bytes: RasaAudioBytes) -> None:
        self.cache[text] = audio_bytes
        self.cache.move_to_end(text)
        if len(self.cache) > self.max_size:
            self.cache.popitem(last=False)



import asyncio
import contextlib
import hashlib
import inspect
import json
import logging
import multiprocessing
import os
import platform
import sys
import textwrap
import typing
import uuid
from collections import defaultdict
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Text, Tuple, Type, cast

import importlib_resources
import requests
import structlog
from terminaltables import SingleTable

import rasa
import rasa.anonymization.utils
import rasa.shared.utils.io
import rasa.utils.io
from rasa import model
from rasa.constants import (
    CONFIG_FILE_TELEMETRY_KEY,
    CONFIG_TELEMETRY_DATE,
    CONFIG_TELEMETRY_ENABLED,
    CONFIG_TELEMETRY_ID,
)
from rasa.engine.storage.local_model_storage import LocalModelStorage
from rasa.shared.constants import (
    ASSISTANT_ID_KEY,
    CONFIG_LANGUAGE_KEY,
    CONFIG_PIPELINE_KEY,
    CONFIG_POLICIES_KEY,
    CONFIG_PREDICT_SCHEMA,
    CONFIG_RECIPE_KEY,
    CONFIG_TRAIN_SCHEMA,
    DOCS_URL_TELEMETRY,
    LLM_API_HEALTH_CHECK_DEFAULT_VALUE,
    LLM_API_HEALTH_CHECK_ENV_VAR,
    MODEL_GROUP_CONFIG_KEY,
    PROMPT_CONFIG_KEY,
    PROMPT_TEMPLATE_CONFIG_KEY,
    UTTER_ASK_PREFIX,
)
from rasa.shared.core.flows import Flow
from rasa.shared.core.flows.steps import (
    CallFlowStep,
    CollectInformationFlowStep,
    LinkFlowStep,
    SetSlotsFlowStep,
)
from rasa.shared.exceptions import RasaException
from rasa.utils import common as rasa_utils

if typing.TYPE_CHECKING:
    from rasa.core.agent import Agent
    from rasa.core.brokers.broker import EventBroker
    from rasa.core.channels.channel import InputChannel
    from rasa.core.tracker_store import TrackerStore
    from rasa.core.utils import AvailableEndpoints
    from rasa.e2e_test.e2e_test_case import Fixture, Metadata, TestCase
    from rasa.shared.importers.importer import TrainingDataImporter
    from rasa.shared.nlu.training_data.training_data import TrainingData

logger = logging.getLogger(__name__)

SEGMENT_TRACK_ENDPOINT = "https://api.segment.io/v1/track"
SEGMENT_IDENTIFY_ENDPOINT = "https://api.segment.io/v1/identify"
SEGMENT_REQUEST_TIMEOUT = 5  # seconds

TELEMETRY_ENABLED_ENVIRONMENT_VARIABLE = "RASA_TELEMETRY_ENABLED"
TELEMETRY_DEBUG_ENVIRONMENT_VARIABLE = "RASA_TELEMETRY_DEBUG"
RASA_PRO_CONFIG_FILE_TELEMETRY_KEY = "traits"

# the environment variable can be used for local development to set a test key
# e.g. `RASA_TELEMETRY_WRITE_KEY=12354 rasa train`
TELEMETRY_WRITE_KEY_ENVIRONMENT_VARIABLE = "RASA_TELEMETRY_WRITE_KEY"
EXCEPTION_WRITE_KEY_ENVIRONMENT_VARIABLE = "RASA_EXCEPTION_WRITE_KEY"

TELEMETRY_ID = "metrics_id"
TELEMETRY_ENABLED_BY_DEFAULT = True

# if one of these environment variables is set, we assume to be running in CI env
CI_ENVIRONMENT_TELL = [
    "bamboo.buildKey",
    "BUILD_ID",
    "BUILD_NUMBER",
    "BUILDKITE",
    "CI",
    "CIRCLECI",
    "CONTINUOUS_INTEGRATION",
    "GITHUB_ACTIONS",
    "HUDSON_URL",
    "JENKINS_URL",
    "TEAMCITY_VERSION",
    "TRAVIS",
    "CODEBUILD_BUILD_ARN",
    "CODEBUILD_BUILD_ID",
    "CODEBUILD_BATCH_BUILD_IDENTIFIER",
]

# If updating or creating a new event, remember to update
# https://rasa.com/docs/rasa-pro/telemetry/telemetry OR
# https://rasa.com/docs/rasa-pro/telemetry/reference
TRAINING_STARTED_EVENT = "Training Started"
TRAINING_COMPLETED_EVENT = "Training Completed"
TRAINING_FAILED_EVENT = "Training Failed"
TELEMETRY_DISABLED_EVENT = "Telemetry Disabled"
TELEMETRY_DATA_SPLIT_EVENT = "Training Data Split"
TELEMETRY_DATA_VALIDATED_EVENT = "Training Data Validated"
TELEMETRY_DATA_CONVERTED_EVENT = "Training Data Converted"
TELEMETRY_TRACKER_EXPORTED_EVENT = "Tracker Exported"
TELEMETRY_INTERACTIVE_LEARNING_STARTED_EVENT = "Interactive Learning Started"
TELEMETRY_SERVER_STARTED_EVENT = "Server Started"
TELEMETRY_PROJECT_CREATED_EVENT = "Project Created"
TELEMETRY_SHELL_STARTED_EVENT = "Shell Started"
TELEMETRY_INSPECT_STARTED_EVENT = "Inspect Started"
TELEMETRY_VISUALIZATION_STARTED_EVENT = "Story Visualization Started"
TELEMETRY_TEST_CORE_EVENT = "Model Core Tested"
TELEMETRY_TEST_NLU_EVENT = "Model NLU Tested"
TELEMETRY_MARKERS_EXTRACTION_INITIATED_EVENT = "Markers Extraction Initiated"
TELEMETRY_MARKERS_EXTRACTED_EVENT = "Markers Extracted"
TELEMETRY_MARKERS_STATS_COMPUTED_EVENT = "Markers Statistics Computed"
TELEMETRY_MARKERS_PARSED_COUNT = "Markers Parsed"

TELEMETRY_RESPONSE_REPHRASED_EVENT = "Response Rephrased"
TELEMETRY_INTENTLESS_POLICY_TRAINING_STARTED_EVENT = (
    "Intentless Policy Training Started"
)
TELEMETRY_INTENTLESS_POLICY_TRAINING_COMPLETED_EVENT = (
    "Intentless Policy Training Completed"
)
TELEMETRY_INTENTLESS_POLICY_PREDICT_EVENT = "Intentless Policy Predicted"
TELEMETRY_E2E_TEST_RUN_STARTED_EVENT = "E2E Test Run Started"
TELEMETRY_ENTERPRISE_SEARCH_POLICY_TRAINING_STARTED_EVENT = (
    "Enterprise Search Policy Training Started"
)
TELEMETRY_ENTERPRISE_SEARCH_POLICY_TRAINING_COMPLETED_EVENT = (
    "Enterprise Search Policy Training Completed"
)
TELEMETRY_ENTERPRISE_SEARCH_POLICY_PREDICT_EVENT = "Enterprise Search Policy Predicted"
TELEMETRY_VALIDATION_ERROR_LOG_EVENT = "Validation Error Logged"
TELEMETRY_UPLOAD_TO_STUDIO_FAILED_EVENT = "Upload to Studio Failed"

# licensing events
TELEMETRY_CONVERSATION_COUNT = "Conversation Count"
TELEMETRY_CONVERSATION_SOFT_LIMIT_REACHED = "Conversation Soft Limit Reached"
TELEMETRY_CONVERSATION_HARD_LIMIT_REACHED = "Conversation Hard Limit Reached"

# used to calculate the context on the first call and cache it afterwards
TELEMETRY_CONTEXT = None

# constants used for the training started events
NUM_FLOWS = "num_flows"
NUM_FLOWS_WITH_NLU_TRIGGER = "num_flows_with_nlu_trigger"
NUM_FLOWS_WITH_FLOW_GUARDS = "num_flows_with_flow_guards"
NUM_FLOWS_ALWAYS_INCLUDED_IN_PROMPT = "num_flows_always_included_in_prompt"
NUM_FLOWS_WITH_NOT_STARTABLE_FLOW_GUARDS = "num_flows_with_not_startable_flow_guards"
NUM_COLLECT_STEPS = "num_collect_steps"
NUM_COLLECT_STEPS_WITH_SEPARATE_UTTER = "num_collect_steps_with_separate_utter"
NUM_COLLECT_STEPS_WITH_REJECTIONS = "num_collect_steps_with_rejections"
NUM_COLLECT_STEPS_WITH_NOT_RESET_AFTER_FLOW_ENDS = (
    "num_collect_steps_with_not_reset_after_flow_ends"
)
NUM_SET_SLOT_STEPS = "num_set_slot_steps"
MAX_DEPTH_OF_IF_CONSTRUCT = "max_depth_of_if_construct"
NUM_LINK_STEPS = "num_link_steps"
NUM_CALL_STEPS = "num_call_steps"
NUM_SHARED_SLOTS_BETWEEN_FLOWS = "num_shared_slots_between_flows"
LLM_COMMAND_GENERATOR_MODEL_NAME = "llm_command_generator_model_name"
LLM_COMMAND_GENERATOR_MODEL_GROUP_ID = "llm_command_generator_model_group_id"
LLM_COMMAND_GENERATOR_CUSTOM_PROMPT_USED = "llm_command_generator_custom_prompt_used"
MULTI_STEP_LLM_COMMAND_GENERATOR_HANDLE_FLOWS_PROMPT_USED = (
    "multi_step_llm_command_generator_custom_handle_flows_prompt_used"
)
MULTI_STEP_LLM_COMMAND_GENERATOR_FILL_SLOTS_PROMPT_USED = (
    "multi_step_llm_command_generator_custom_fill_slots_prompt_used"
)
FLOW_RETRIEVAL_ENABLED = "flow_retrieval_enabled"
FLOW_RETRIEVAL_EMBEDDING_MODEL_NAME = "flow_retrieval_embedding_model_name"
FLOW_RETRIEVAL_EMBEDDING_MODEL_GROUP_ID = "flow_retrieval_embedding_model_group_id"
TRACING_BACKEND = "tracing_backend"
METRICS_BACKEND = "metrics_backend"
VERSION = "version"

# E2E test conversion
TELEMETRY_E2E_TEST_CONVERSION_EVENT = "E2E Test Conversion Completed"
E2E_TEST_CONVERSION_FILE_TYPE = "file_type"
E2E_TEST_CONVERSION_TEST_CASE_COUNT = "test_case_count"


def print_telemetry_reporting_info() -> None:
    """Print telemetry information to std out."""
    message = textwrap.dedent(
        f"""
        Rasa Pro reports anonymous usage telemetry to help improve the product
        for all its users.

        If you'd like to opt-out, you can use `rasa telemetry disable`.
        To learn more, check out {DOCS_URL_TELEMETRY}."""
    ).strip()

    table = SingleTable([[message]])
    print(table.table)


def _default_telemetry_configuration(is_enabled: bool) -> Dict[Text, Any]:
    return {
        CONFIG_TELEMETRY_ENABLED: is_enabled,
        CONFIG_TELEMETRY_ID: uuid.uuid4().hex,
        CONFIG_TELEMETRY_DATE: datetime.now(),
    }


def _write_default_telemetry_configuration(
    is_enabled: bool = TELEMETRY_ENABLED_BY_DEFAULT,
) -> bool:
    new_config = _default_telemetry_configuration(is_enabled)

    keys = [CONFIG_FILE_TELEMETRY_KEY, RASA_PRO_CONFIG_FILE_TELEMETRY_KEY]

    success = all(
        [rasa_utils.write_global_config_value(key, new_config) for key in keys]
    )

    # Do not show info if user has enabled/disabled telemetry via env var
    telemetry_environ = os.environ.get(TELEMETRY_ENABLED_ENVIRONMENT_VARIABLE)
    if is_enabled and success and telemetry_environ is None:
        print_telemetry_reporting_info()

    return success


def _is_telemetry_enabled_in_configuration() -> bool:
    """Read telemetry configuration from the user's Rasa config file in $HOME.

    Creates a default configuration if no configuration exists.

    Returns:
        `True`, if telemetry is enabled, `False` otherwise.
    """
    try:
        stored_config = rasa_utils.read_global_config_value(
            CONFIG_FILE_TELEMETRY_KEY, unavailable_ok=False
        )

        return stored_config[CONFIG_TELEMETRY_ENABLED]
    except ValueError as e:
        logger.debug(f"Could not read telemetry settings from configuration file: {e}")

        # seems like there is no config, we'll create one and enable telemetry
        success = _write_default_telemetry_configuration()
        # if writing the configuration failed, telemetry will be disabled
        return TELEMETRY_ENABLED_BY_DEFAULT and success


def is_telemetry_enabled() -> bool:
    """Check if telemetry is enabled either in configuration or environment.

    Returns:
        `True`, if telemetry is enabled, `False` otherwise.
    """
    from rasa.utils import licensing

    if licensing.is_champion_server_license():
        logger.debug("Telemetry is enabled for developer licenses.")
        return True

    telemetry_environ = os.environ.get(TELEMETRY_ENABLED_ENVIRONMENT_VARIABLE)

    if telemetry_environ is not None:
        return telemetry_environ.lower() == "true"

    try:
        return rasa_utils.read_global_config_value(
            CONFIG_FILE_TELEMETRY_KEY, unavailable_ok=False
        )[CONFIG_TELEMETRY_ENABLED]
    except ValueError:
        return False


def initialize_telemetry() -> bool:
    """Read telemetry configuration from the user's Rasa config file in $HOME.

    Creates a default configuration if no configuration exists.

    Returns:
        `True`, if telemetry is enabled, `False` otherwise.
    """
    try:
        # calling this even if the environment variable is set makes sure the
        # configuration is created and there is a telemetry ID
        is_enabled_in_configuration = _is_telemetry_enabled_in_configuration()

        telemetry_environ = os.environ.get(TELEMETRY_ENABLED_ENVIRONMENT_VARIABLE)

        if telemetry_environ is None:
            return is_enabled_in_configuration

        return telemetry_environ.lower() == "true"
    except Exception as e:  # skipcq:PYL-W0703
        logger.exception(
            f"Failed to initialize telemetry reporting: {e}."
            f"Telemetry reporting will be disabled."
        )
        return False


def ensure_telemetry_enabled(f: Callable[..., Any]) -> Callable[..., Any]:
    """Function decorator for telemetry functions that ensures telemetry is enabled.

    WARNING: does not work as a decorator for async generators.

    Args:
        f: function to call if telemetry is enabled
    Returns:
        Return wrapped function
    """
    # allows us to use the decorator for async generator functions
    if inspect.isasyncgenfunction(f):

        @wraps(f)
        async def decorated_async_gen(*args: Any, **kwargs: Any) -> Any:
            if is_telemetry_enabled():
                yield f(*args, **kwargs)

        return decorated_async_gen

    # allows us to use the decorator for async and non async functions
    if asyncio.iscoroutinefunction(f):

        @wraps(f)
        async def decorated_coroutine(*args: Any, **kwargs: Any) -> Any:
            if is_telemetry_enabled():
                return await f(*args, **kwargs)
            return None

        return decorated_coroutine

    @wraps(f)
    def decorated(*args: Any, **kwargs: Any) -> Any:
        if is_telemetry_enabled():
            return f(*args, **kwargs)
        return None

    return decorated


def _fetch_write_key(tool: Text, environment_variable: Text) -> Optional[Text]:
    """Read the write key from a tool from our set of keys.

    Args:
        tool: name of the tool we want to fetch a key for
        environment_variable: name of the environment variable to set the key
    Returns:
        write key, if a key was present.
    """
    import importlib_resources

    from rasa import __name__ as name

    if os.environ.get(environment_variable):
        # a write key set using the environment variable will always
        # overwrite any key provided as part of the package (`keys` file)
        return os.environ.get(environment_variable)

    write_key_path = str(importlib_resources.files(name).joinpath("keys"))

    # noinspection PyBroadException
    try:
        with open(write_key_path) as f:
            return json.load(f).get(tool)
    except Exception:  # skipcq:PYL-W0703
        return None


def telemetry_write_key() -> Optional[Text]:
    """Read the Segment write key from the segment key text file.

    The segment key text file should by present only in wheel/sdist packaged
    versions of Rasa Pro. This avoids running telemetry locally when
    developing on Rasa or when running CI builds.

    In local development, this should always return `None` to avoid logging telemetry.

    Returns:
        Segment write key, if the key file was present.
    """
    return _fetch_write_key("segment", TELEMETRY_WRITE_KEY_ENVIRONMENT_VARIABLE)


def sentry_write_key() -> Optional[Text]:
    """Read the sentry write key from the sentry key text file.

    Returns:
        Sentry write key, if the key file was present.
    """
    return _fetch_write_key("sentry", EXCEPTION_WRITE_KEY_ENVIRONMENT_VARIABLE)


def _encode_base64(original: Text, encoding: Text = "utf-8") -> Text:
    """Encodes a string as a base64 string.

    Args:
        original: Text to be encoded.
        encoding: Encoding used to convert text to binary.

    Returns:
        Encoded text.
    """
    import base64

    return base64.b64encode(original.encode(encoding)).decode(encoding)


def segment_request_header(write_key: Text) -> Dict[Text, Any]:
    """Use a segment write key to create authentication headers for the segment API.

    Args:
        write_key: Authentication key for segment.

    Returns:
        Authentication headers for segment.
    """
    return {
        "Authorization": "Basic {}".format(_encode_base64(write_key + ":")),
        "Content-Type": "application/json",
    }


def segment_request_payload(
    distinct_id: Text,
    event_name: Text,
    properties: Dict[Text, Any],
    context: Dict[Text, Any],
) -> Dict[Text, Any]:
    """Compose a valid payload for the segment API.

    Args:
        distinct_id: Unique telemetry ID.
        event_name: Name of the event.
        properties: Values to report along the event.
        context: Context information about the event.

    Returns:
        Valid segment payload.
    """
    return {
        "userId": distinct_id,
        "event": event_name,
        "properties": properties,
        "context": context,
    }


def in_continuous_integration() -> bool:
    """Returns `True` if currently running inside a continuous integration context."""
    return any(env in os.environ for env in CI_ENVIRONMENT_TELL)


def _is_telemetry_debug_enabled() -> bool:
    """Check if telemetry debug mode is enabled."""
    return (
        os.environ.get(TELEMETRY_DEBUG_ENVIRONMENT_VARIABLE, "false").lower() == "true"
    )


def print_telemetry_payload(payload: Dict[Text, Any]) -> None:
    """Print a telemetry payload to the commandline.

    Args:
        payload: payload to be delivered to segment.
    """
    payload_json = json.dumps(payload, indent=2)
    logger.debug(f"Telemetry payload: {payload_json}")


def _get_telemetry_write_key() -> Optional[Text]:
    if os.environ.get(TELEMETRY_WRITE_KEY_ENVIRONMENT_VARIABLE):
        return os.environ.get(TELEMETRY_WRITE_KEY_ENVIRONMENT_VARIABLE)

    write_key_path = str(importlib_resources.files(rasa.__name__).joinpath("keys"))

    try:
        with open(write_key_path) as f:
            return json.load(f).get("segment")
    except Exception:
        return None


def _send_event(
    distinct_id: Text,
    event_name: Text,
    properties: Dict[Text, Any],
    context: Dict[Text, Any],
) -> None:
    """Report the contents segmentof an event to the /track Segment endpoint.

    Documentation: https://.com/docs/sources/server/http/

    Do not call this function from outside telemetry.py! This function does not
    check if telemetry is enabled or not.

    Args:
        distinct_id: Unique telemetry ID.
        event_name: Name of the event.
        properties: Values to report along the event.
        context: Context information about the event.
    """
    payload = segment_request_payload(distinct_id, event_name, properties, context)

    _send_request(SEGMENT_TRACK_ENDPOINT, payload)


def _send_request(url: Text, payload: Dict[Text, Any]) -> None:
    """Send a request to the Segment API.

    Args:
        url: URL of the Segment API endpoint
        payload: payload to send to the Segment API
    """
    if _is_telemetry_debug_enabled():
        print_telemetry_payload(payload)

    write_key = _get_telemetry_write_key()
    if not write_key:
        # If RASA_TELEMETRY_WRITE_KEY is empty or `None`, telemetry has not
        # been enabled for this build (e.g. because it is running from source)
        logger.debug("Skipping request to external service: telemetry key not set.")
        return

    headers = rasa.telemetry.segment_request_header(write_key)

    resp = requests.post(
        url=url,
        headers=headers,
        json=payload,
        timeout=SEGMENT_REQUEST_TIMEOUT,
    )
    # handle different failure cases
    if resp.status_code != 200:
        logger.debug(
            f"Segment telemetry request returned a {resp.status_code} response. "
            f"Body: {resp.text}"
        )
    else:
        data = resp.json()
        if not data.get("success"):
            logger.debug(
                f"Segment telemetry request returned a failure. Response: {data}"
            )


def _hash_directory_path(path: Text) -> Optional[Text]:
    """Create a hash for the directory.

    Returns:
        hash of the directories path
    """
    full_path = Path(path).absolute()
    return hashlib.sha256(str(full_path).encode("utf-8")).hexdigest()


# noinspection PyBroadException
def _is_docker() -> bool:
    """Guess if we are running in docker environment.

    Returns:
        `True` if we are running inside docker, `False` otherwise.
    """
    # first we try to use the env
    try:
        os.stat("/.dockerenv")
        return True
    except Exception:  # skipcq:PYL-W0703
        pass

    # if that didn't work, try to use proc information
    try:
        return "docker" in rasa.shared.utils.io.read_file("/proc/self/cgroup", "utf8")
    except Exception:  # skipcq:PYL-W0703
        return False


def with_default_context_fields(
    context: Optional[Dict[Text, Any]] = None,
) -> Dict[Text, Any]:
    """Return a new context dictionary with default and provided field values merged.

    The default fields contain only the OS information for now.

    Args:
        context: Context information about the event.

    Return:
        A new context.
    """
    context = context or {}

    return {**_default_context_fields(), **context}


def _default_context_fields() -> Dict[Text, Any]:
    """Return a dictionary that contains the default context values.

    Return:
        A new context containing information about the runtime environment.
    """
    from rasa.utils.licensing import get_license_hash, property_of_active_license

    global TELEMETRY_CONTEXT

    if not TELEMETRY_CONTEXT:
        # Make sure to update the example in docs/docs/telemetry/telemetry.mdx
        # if you change / add context
        TELEMETRY_CONTEXT = {
            "os": {"name": platform.system(), "version": platform.release()},
            "ci": in_continuous_integration(),
            "project": model.project_fingerprint(),
            "directory": _hash_directory_path(os.getcwd()),
            "python": sys.version.split(" ")[0],
            "rasa_pro": rasa.__version__,
            "cpu": multiprocessing.cpu_count(),
            "docker": _is_docker(),
            "license_hash": get_license_hash(),
            "company": property_of_active_license(
                lambda active_license: active_license.company
            ),
        }

    # avoid returning the cached dict --> caller could modify the dictionary...
    # usually we would use `lru_cache`, but that doesn't return a dict copy and
    # doesn't work on inner functions, so we need to roll our own caching...
    return TELEMETRY_CONTEXT.copy()


def _track(
    event_name: Text,
    properties: Optional[Dict[Text, Any]] = None,
    context: Optional[Dict[Text, Any]] = None,
) -> None:
    """Tracks a telemetry event.

    It is OK to use this function from outside telemetry.py, but note that it
    is recommended to create a new track_xyz() function for complex telemetry
    events, or events that are generated from many parts of the Rasa Pro code.

    Args:
        event_name: Name of the event.
        properties: Dictionary containing the event's properties.
        context: Dictionary containing some context for this event.
    """
    try:
        telemetry_id = get_telemetry_id()

        if not telemetry_id:
            logger.debug("Will not report telemetry events as no ID was found.")
            return

        if not properties:
            properties = {}

        properties[TELEMETRY_ID] = telemetry_id

        # this is an additional check in case _track() is called
        # from a function that is not decorated with @ensure_telemetry_enabled
        if is_telemetry_enabled():
            _send_event(
                telemetry_id,
                event_name,
                properties,
                with_default_context_fields(context),
            )
    except Exception as e:  # skipcq:PYL-W0703
        logger.debug(f"Skipping telemetry reporting: {e}")


def _identify(
    traits: Optional[Dict[Text, Any]] = None,
    context: Optional[Dict[Text, Any]] = None,
) -> None:
    """Tracks telemetry traits.

    It is OK to use this function from outside telemetry.py, but note that it
    is recommended to create a new track_xyz() function for complex telemetry
    traits, or traits that are generated from many parts of the Rasa Pro code.

    Args:
        traits: Dictionary containing the tracked traits.
        context: Dictionary containing some context for the traits.
    """
    try:
        telemetry_id = get_telemetry_id()

        if not telemetry_id:
            logger.debug("Will not report telemetry events as no ID was found.")
            return

        if not traits:
            traits = {}

        _send_traits(telemetry_id, traits, with_default_context_fields(context))
    except Exception as e:
        logger.debug(f"Skipping telemetry reporting: {e}")


def _send_traits(
    distinct_id: Text,
    traits: Dict[Text, Any],
    context: Dict[Text, Any],
) -> None:
    """Report the contents of telemetry traits to the /identify Segment endpoint.

    Do not call this function from outside telemetry.py! This function does not
    check if telemetry is enabled or not.

    Args:
        distinct_id: Unique telemetry ID.
        traits: Pieces of information to be recorded about
                rasa_plus interface implementations.
        context: Context information to be sent along with traits.
    """
    payload = segment_identify_request_payload(distinct_id, traits, context)

    _send_request(SEGMENT_IDENTIFY_ENDPOINT, payload)


def segment_identify_request_payload(
    distinct_id: Text,
    traits: Dict[Text, Any],
    context: Dict[Text, Any],
) -> Dict[Text, Any]:
    """Compose a valid payload for the segment API.

    Args:
        distinct_id: Unique telemetry ID.
        traits: Pieces of information to be recorded about
                rasa_plus interface implementations.
        context: Context information to be sent along with traits.

    Returns:
        Valid segment payload.
    """
    return {
        "userId": distinct_id,
        "traits": traits,
        "context": context,
    }


def get_telemetry_id() -> Optional[Text]:
    """Return the unique telemetry identifier for this Rasa Pro install.

    The identifier can be based on the license.
    Otherwise, it can be any string, but it should be a UUID.

    Returns:
        The identifier, if it is configured correctly.
    """
    from rasa.utils.licensing import property_of_active_license

    return property_of_active_license(lambda active_license: active_license.jti)


def toggle_telemetry_reporting(is_enabled: bool) -> None:
    """Write to the configuration if telemetry tracking should be enabled or disabled.

    Args:
        is_enabled: `True` if the telemetry reporting should be enabled,
            `False` otherwise.
    """
    configuration = rasa_utils.read_global_config_value(CONFIG_FILE_TELEMETRY_KEY)

    if configuration:
        configuration[CONFIG_TELEMETRY_ENABLED] = is_enabled
    else:
        configuration = _default_telemetry_configuration(is_enabled)

    rasa_utils.write_global_config_value(CONFIG_FILE_TELEMETRY_KEY, configuration)
    rasa_utils.write_global_config_value(
        RASA_PRO_CONFIG_FILE_TELEMETRY_KEY, configuration
    )


def filter_errors(
    event: Optional[Dict[Text, Any]], hint: Optional[Dict[Text, Any]] = None
) -> Optional[Dict[Text, Any]]:
    """Filter errors.

    Args:
        event: event to be logged to sentry
        hint: some hinting information sent alongside of the event

    Returns:
        the event without any sensitive / PII data or `None` if the event constitutes
        an `ImportError` which should be discarded.
    """
    if hint and "exc_info" in hint:
        exc_type, exc_value, tb = hint["exc_info"]
        if isinstance(exc_value, ImportError):
            return None
    return event


def before_send(
    event: Dict[Text, Any], _unused_hint: Optional[Dict[Text, Any]] = None
) -> Optional[Dict[Text, Any]]:
    """Strips the sensitive data and filters errors before sending to sentry.

    Args:
        event: event to be logged to sentry
        _unused_hint: some hinting information sent alongside of the event

    Returns:
        the event without any sensitive / PII data or `None` if the event should
        be discarded.
    """
    cleaned_event = strip_sensitive_data_from_sentry_event(event, _unused_hint)
    return filter_errors(cleaned_event, _unused_hint)


def strip_sensitive_data_from_sentry_event(
    event: Dict[Text, Any], _unused_hint: Optional[Dict[Text, Any]] = None
) -> Optional[Dict[Text, Any]]:
    """Remove any sensitive data from the event (e.g. path names).

    Args:
        event: event to be logged to sentry
        _unused_hint: some hinting information sent alongside of the event

    Returns:
        the event without any sensitive / PII data or `None` if the event should
        be discarded.
    """
    # removes any paths from stack traces (avoids e.g. sending
    # a users home directory name if package is installed there)
    for value in event.get("exception", {}).get("values", []):
        for frame in value.get("stacktrace", {}).get("frames", []):
            frame["abs_path"] = ""

            if f"rasa_sdk{os.path.sep}executor.py" in frame["filename"]:
                # this looks a lot like an exception in the SDK and hence custom code
                # no need for us to deal with that
                return None
            elif "site-packages" in frame["filename"]:
                # drop site-packages and following slash / backslash
                relative_name = frame["filename"].split("site-packages")[-1][1:]
                frame["filename"] = os.path.join("site-packages", relative_name)
            elif "dist-packages" in frame["filename"]:
                # drop dist-packages and following slash / backslash
                relative_name = frame["filename"].split("dist-packages")[-1][1:]
                frame["filename"] = os.path.join("dist-packages", relative_name)
            elif os.path.isabs(frame["filename"]):
                # if the file path is absolute, we'll drop the whole event as this is
                # very likely custom code. needs to happen after cleaning as
                # site-packages / dist-packages paths are also absolute, but fine.
                return None
    return event


@ensure_telemetry_enabled
def initialize_error_reporting() -> None:
    """Sets up automated error reporting.

    Exceptions are reported to sentry. We avoid sending any metadata (local
    variables, paths, ...) to make sure we don't compromise any data. Only the
    exception and its stacktrace is logged and only if the exception origins
    from the `rasa` package.
    """
    import sentry_sdk
    from sentry_sdk import configure_scope
    from sentry_sdk.integrations.atexit import AtexitIntegration
    from sentry_sdk.integrations.dedupe import DedupeIntegration
    from sentry_sdk.integrations.excepthook import ExcepthookIntegration

    # key for local testing can be found at
    # https://sentry.io/settings/rasahq/projects/rasa-open-source/install/python/
    # for local testing, set the key using `RASA_EXCEPTION_WRITE_KEY=key rasa <command>`
    key = sentry_write_key()

    if not key:
        return

    telemetry_id = get_telemetry_id()

    # this is a very defensive configuration, avoiding as many integrations as
    # possible. it also submits very little data (exception with error message
    # and line numbers).
    sentry_sdk.init(
        f"https://{key}.ingest.sentry.io/2801673",
        before_send=before_send,
        integrations=[
            ExcepthookIntegration(),
            DedupeIntegration(),
            AtexitIntegration(lambda _, __: None),
        ],
        send_default_pii=False,  # activate PII filter
        server_name=telemetry_id or "UNKNOWN",
        ignore_errors=[
            # std lib errors
            KeyboardInterrupt,  # user hit the interrupt key (Ctrl+C)
            MemoryError,  # machine is running out of memory
            NotImplementedError,  # user is using a feature that is not implemented
            asyncio.CancelledError,  # an async operation has been cancelled by the user
            # expected Rasa errors
            RasaException,
            OSError,
        ],
        in_app_include=["rasa"],  # only submit errors in this package
        include_local_variables=False,  # don't submit local variables
        release=f"rasa-{rasa.__version__}",
        default_integrations=False,
        environment="development" if in_continuous_integration() else "production",
    )

    if not telemetry_id:
        return

    with configure_scope() as scope:
        # sentry added these more recently, just a protection in a case where a
        # user has installed an older version of sentry
        if hasattr(scope, "set_user"):
            scope.set_user({"id": telemetry_id})

        default_context = _default_context_fields()
        if hasattr(scope, "set_context"):
            if "os" in default_context:
                # os is a nested dict, hence we report it separately
                scope.set_context("Operating System", default_context.pop("os"))
            scope.set_context("Environment", default_context)


@contextlib.contextmanager
def track_model_training(
    training_data: "TrainingDataImporter", model_type: Text, is_finetuning: bool = False
) -> typing.Generator[None, None, None]:
    """Track a model training started.

    WARNING: since this is a generator, it can't use the ensure telemetry
        decorator. We need to manually add these checks here. This can be
        fixed as soon as we drop python 3.6 support.

    Args:
        training_data: Training data used for the training.
        model_type: Specifies the type of training, should be either "rasa", "core"
            or "nlu".
        is_finetuning: `True` if the model is trained by finetuning another model.
    """
    if not initialize_telemetry():
        # telemetry reporting is disabled. we won't do any reporting
        yield  # runs the training
        return

    config = training_data.get_config()
    stories = training_data.get_stories()
    nlu_data = training_data.get_nlu_data()
    domain = training_data.get_domain()
    flows = training_data.get_flows()
    count_conditional_responses = domain.count_conditional_response_variations()
    (
        count_total_mappings,
        count_custom_mappings,
        count_conditional_mappings,
    ) = domain.count_slot_mapping_statistics()

    training_id = uuid.uuid4().hex

    tracking_data = {
        "language": config.get(CONFIG_LANGUAGE_KEY),
        "training_id": training_id,
        "type": model_type,
        "assistant_id": config.get(ASSISTANT_ID_KEY),
        "pipeline": config.get(CONFIG_PIPELINE_KEY),
        "policies": config.get(CONFIG_POLICIES_KEY),
        "train_schema": config.get(CONFIG_TRAIN_SCHEMA),
        "predict_schema": config.get(CONFIG_PREDICT_SCHEMA),
        "model_groups": rasa.core.utils.AvailableEndpoints.get_instance().model_groups,
        "api_health_check_enabled": (
            os.getenv(
                LLM_API_HEALTH_CHECK_ENV_VAR, LLM_API_HEALTH_CHECK_DEFAULT_VALUE
            ).lower()
            == "true"
        ),
        "num_intent_examples": len(nlu_data.intent_examples),
        "num_entity_examples": len(nlu_data.entity_examples),
        "num_actions": len(domain.action_names_or_texts),
        # Old nomenclature from when 'responses' were still called
        # 'templates' in the domain
        "num_templates": len(domain.responses),
        "num_conditional_response_variations": count_conditional_responses,
        "num_slot_mappings": count_total_mappings,
        "num_custom_slot_mappings": count_custom_mappings,
        "num_conditional_slot_mappings": count_conditional_mappings,
        "num_slots": len(domain.slots),
        "num_forms": len(domain.forms),
        "num_intents": len(domain.intents),
        "num_entities": len(domain.entities),
        "num_story_steps": len(stories.story_steps),
        "num_lookup_tables": len(nlu_data.lookup_tables),
        "num_synonyms": len(nlu_data.entity_synonyms),
        "num_regexes": len(nlu_data.regex_features),
        "is_finetuning": is_finetuning,
        "recipe": config.get(CONFIG_RECIPE_KEY),
    }

    flow_statistics = _collect_flow_statistics(flows.underlying_flows)
    tracking_data.update(flow_statistics)
    command_generator_settings = _get_llm_command_generator_config(config)
    tracking_data.update(command_generator_settings)

    # Make sure to update the example in docs/docs/telemetry/telemetry.mdx
    # if you change / add any properties
    _track(
        TRAINING_STARTED_EVENT,
        tracking_data,
    )
    start = datetime.now()
    try:
        yield
    except (Exception, SystemExit):
        runtime = datetime.now() - start
        _track(
            TRAINING_FAILED_EVENT,
            {
                "training_id": training_id,
                "assistant_id": config.get(ASSISTANT_ID_KEY),
                "type": model_type,
                "runtime": int(runtime.total_seconds()),
            },
        )
        raise

    runtime = datetime.now() - start

    _track(
        TRAINING_COMPLETED_EVENT,
        {
            "training_id": training_id,
            "assistant_id": config.get(ASSISTANT_ID_KEY),
            "type": model_type,
            "runtime": int(runtime.total_seconds()),
        },
    )


def _collect_flow_statistics(flows: List[Flow]) -> Dict[str, Any]:
    """Collects some statistics about the flows, such as number of specific steps."""
    data = {
        NUM_FLOWS: len(flows),
        NUM_FLOWS_WITH_NLU_TRIGGER: 0,
        NUM_FLOWS_WITH_FLOW_GUARDS: 0,
        NUM_FLOWS_ALWAYS_INCLUDED_IN_PROMPT: 0,
        NUM_FLOWS_WITH_NOT_STARTABLE_FLOW_GUARDS: 0,
        NUM_COLLECT_STEPS: 0,
        NUM_COLLECT_STEPS_WITH_SEPARATE_UTTER: 0,
        NUM_COLLECT_STEPS_WITH_REJECTIONS: 0,
        NUM_COLLECT_STEPS_WITH_NOT_RESET_AFTER_FLOW_ENDS: 0,
        NUM_SET_SLOT_STEPS: 0,
        MAX_DEPTH_OF_IF_CONSTRUCT: 0,
        NUM_LINK_STEPS: 0,
        NUM_CALL_STEPS: 0,
        NUM_SHARED_SLOTS_BETWEEN_FLOWS: 0,
    }

    slots_used_in_different_flows = defaultdict(set)

    for flow in flows:
        if flow.guard_condition:
            data[NUM_FLOWS_WITH_FLOW_GUARDS] += 1
            if flow.guard_condition.lower() == "false":
                data[NUM_FLOWS_WITH_NOT_STARTABLE_FLOW_GUARDS] += 1

        if flow.always_include_in_prompt:
            data[NUM_FLOWS_ALWAYS_INCLUDED_IN_PROMPT] += 1

        if flow.nlu_triggers:
            data[NUM_FLOWS_WITH_NLU_TRIGGER] += 1

        for step in flow.steps_with_calls_resolved:
            if isinstance(step, CollectInformationFlowStep):
                slots_used_in_different_flows[step.collect].add(flow.id)
                data[NUM_COLLECT_STEPS] += 1
                if len(step.rejections) > 0:
                    data[NUM_COLLECT_STEPS_WITH_REJECTIONS] += 1
                if not step.reset_after_flow_ends:
                    data[NUM_COLLECT_STEPS_WITH_NOT_RESET_AFTER_FLOW_ENDS] += 1
                if step.utter != f"{UTTER_ASK_PREFIX}{step.collect}":
                    data[NUM_COLLECT_STEPS_WITH_SEPARATE_UTTER] += 1

            if isinstance(step, SetSlotsFlowStep):
                for slot in step.slots:
                    slots_used_in_different_flows[slot["key"]].add(flow.id)
                data[NUM_SET_SLOT_STEPS] += 1

            if isinstance(step, LinkFlowStep):
                data[NUM_LINK_STEPS] += 1

            if isinstance(step, CallFlowStep):
                data[NUM_CALL_STEPS] += 1

            if step.next:
                depth = step.next.depth_in_tree()
                if depth > data[MAX_DEPTH_OF_IF_CONSTRUCT]:
                    data[MAX_DEPTH_OF_IF_CONSTRUCT] = depth

    for flows_with_slot in slots_used_in_different_flows.values():
        if len(flows_with_slot) > 1:
            data[NUM_SHARED_SLOTS_BETWEEN_FLOWS] += 1

    return data


def _get_llm_command_generator_config(config: Dict[str, Any]) -> Optional[Dict]:
    """Returns the configuration for the LLMCommandGenerator.

    Includes the model name, whether a custom prompt is used, whether flow
    retrieval is enabled, and flow retrieval embedding model.
    """
    from rasa.dialogue_understanding.generator import (
        CompactLLMCommandGenerator,
        LLMBasedCommandGenerator,
        LLMCommandGenerator,
        MultiStepLLMCommandGenerator,
        SingleStepLLMCommandGenerator,
    )
    from rasa.dialogue_understanding.generator.constants import (
        FLOW_RETRIEVAL_KEY,
        LLM_CONFIG_KEY,
    )
    from rasa.dialogue_understanding.generator.flow_retrieval import (
        DEFAULT_EMBEDDINGS_CONFIG,
    )
    from rasa.dialogue_understanding.generator.multi_step.multi_step_llm_command_generator import (  # noqa: E501
        FILL_SLOTS_KEY,
        HANDLE_FLOWS_KEY,
    )
    from rasa.shared.constants import (
        EMBEDDINGS_CONFIG_KEY,
        MODEL_CONFIG_KEY,
        MODEL_NAME_CONFIG_KEY,
    )

    def find_command_generator_component(pipeline: List) -> Optional[Dict]:
        """Finds the LLMCommandGenerator component in the pipeline."""
        for component in pipeline:
            if component["name"] in [
                LLMCommandGenerator.__name__,
                SingleStepLLMCommandGenerator.__name__,
                MultiStepLLMCommandGenerator.__name__,
                CompactLLMCommandGenerator.__name__,
            ]:
                return component
        return None

    def extract_llm_command_generator_llm_client_settings(component: Dict) -> Dict:
        """Extracts settings related to LLM command generator."""
        component_class_lookup = {
            LLMCommandGenerator.__name__: LLMCommandGenerator,
            SingleStepLLMCommandGenerator.__name__: SingleStepLLMCommandGenerator,
            MultiStepLLMCommandGenerator.__name__: MultiStepLLMCommandGenerator,
            CompactLLMCommandGenerator.__name__: CompactLLMCommandGenerator,
        }
        llm_config = component.get(LLM_CONFIG_KEY, {})
        # Config at this stage is not yet resolved, so read from `model_group`
        llm_model_group_id = llm_config.get(MODEL_GROUP_CONFIG_KEY)
        llm_model_name = llm_config.get(MODEL_CONFIG_KEY) or llm_config.get(
            MODEL_NAME_CONFIG_KEY
        )
        if llm_model_group_id is None and llm_model_name is None:
            component_clz = cast(
                Type[LLMBasedCommandGenerator],
                component_class_lookup[component["name"]],
            )
            llm_model_name = component_clz.get_default_llm_config()[MODEL_CONFIG_KEY]

        custom_prompt_used = (
            PROMPT_CONFIG_KEY in component or PROMPT_TEMPLATE_CONFIG_KEY in component
        )
        return {
            LLM_COMMAND_GENERATOR_MODEL_NAME: llm_model_name,
            LLM_COMMAND_GENERATOR_MODEL_GROUP_ID: llm_model_group_id,
            LLM_COMMAND_GENERATOR_CUSTOM_PROMPT_USED: custom_prompt_used,
        }

    def extract_multistep_command_generator_prompt_settings(component: Dict) -> Dict:
        """Extracts settings related to multistep command generator."""
        prompt_templates = component.get("prompt_templates", {})
        handle_flows_prompt_used = HANDLE_FLOWS_KEY in prompt_templates
        fill_slots_prompt_used = FILL_SLOTS_KEY in prompt_templates
        return {
            MULTI_STEP_LLM_COMMAND_GENERATOR_HANDLE_FLOWS_PROMPT_USED: handle_flows_prompt_used,  # noqa: E501
            MULTI_STEP_LLM_COMMAND_GENERATOR_FILL_SLOTS_PROMPT_USED: fill_slots_prompt_used,  # noqa: E501
        }

    def extract_flow_retrieval_settings(component: Dict) -> Dict:
        """Extracts settings related to flow retrieval."""
        flow_retrieval_config = component.get(FLOW_RETRIEVAL_KEY, {})
        flow_retrieval_enabled = flow_retrieval_config.get("active", True)
        embeddings_config = flow_retrieval_config.get(
            EMBEDDINGS_CONFIG_KEY, DEFAULT_EMBEDDINGS_CONFIG
        )
        flow_retrieval_embedding_model_name = (
            (
                embeddings_config.get(MODEL_NAME_CONFIG_KEY)
                or embeddings_config.get(MODEL_CONFIG_KEY)
            )
            if flow_retrieval_enabled
            else None
        )
        # Config at this stage is not yet resolved, so read from `model_group`
        flow_retrieval_embedding_model_group_id = embeddings_config.get(
            MODEL_GROUP_CONFIG_KEY
        )
        return {
            FLOW_RETRIEVAL_ENABLED: flow_retrieval_enabled,
            FLOW_RETRIEVAL_EMBEDDING_MODEL_NAME: flow_retrieval_embedding_model_name,
            FLOW_RETRIEVAL_EMBEDDING_MODEL_GROUP_ID: flow_retrieval_embedding_model_group_id,  # noqa: E501
        }

    def extract_settings(component: Dict) -> Dict:
        """Extracts the settings from the command generator component."""
        settings = {}
        settings.update(extract_llm_command_generator_llm_client_settings(component))
        settings.update(extract_multistep_command_generator_prompt_settings(component))
        settings.update(extract_flow_retrieval_settings(component))
        return settings

    command_generator_config = {
        LLM_COMMAND_GENERATOR_MODEL_NAME: None,
        LLM_COMMAND_GENERATOR_MODEL_GROUP_ID: None,
        LLM_COMMAND_GENERATOR_CUSTOM_PROMPT_USED: None,
        MULTI_STEP_LLM_COMMAND_GENERATOR_HANDLE_FLOWS_PROMPT_USED: None,
        MULTI_STEP_LLM_COMMAND_GENERATOR_FILL_SLOTS_PROMPT_USED: None,
        FLOW_RETRIEVAL_ENABLED: None,
        FLOW_RETRIEVAL_EMBEDDING_MODEL_NAME: None,
        FLOW_RETRIEVAL_EMBEDDING_MODEL_GROUP_ID: None,
    }

    pipeline = config.get(CONFIG_PIPELINE_KEY, [])
    if not isinstance(pipeline, list):
        return command_generator_config

    command_generator_component = find_command_generator_component(pipeline)
    if command_generator_component is not None:
        extracted_settings = extract_settings(command_generator_component)
        command_generator_config.update(extracted_settings)

    return command_generator_config


@ensure_telemetry_enabled
def track_telemetry_disabled() -> None:
    """Track when a user disables telemetry."""
    _track(TELEMETRY_DISABLED_EVENT)


@ensure_telemetry_enabled
def track_data_split(fraction: float, data_type: Text) -> None:
    """Track when a user splits data.

    Args:
        fraction: How much data goes into train and how much goes into test
        data_type: Is this core, nlu or nlg data
    """
    _track(TELEMETRY_DATA_SPLIT_EVENT, {"fraction": fraction, "type": data_type})


@ensure_telemetry_enabled
def track_validate_files(validation_success: bool) -> None:
    """Track when a user validates data files.

    Args:
        validation_success: Whether the validation was successful
    """
    _track(TELEMETRY_DATA_VALIDATED_EVENT, {"validation_success": validation_success})


@ensure_telemetry_enabled
def track_data_convert(output_format: Text, data_type: Text) -> None:
    """Track when a user converts data.

    Args:
        output_format: Target format for the converter
        data_type: Is this core, nlu or nlg data
    """
    _track(
        TELEMETRY_DATA_CONVERTED_EVENT,
        {"output_format": output_format, "type": data_type},
    )


@ensure_telemetry_enabled
def track_tracker_export(
    number_of_exported_events: int,
    tracker_store: "TrackerStore",
    event_broker: "EventBroker",
) -> None:
    """Track when a user exports trackers.

    Args:
        number_of_exported_events: Number of events that got exported
        tracker_store: Store used to retrieve the events from
        event_broker: Broker the events are getting published towards
    """
    _track(
        TELEMETRY_TRACKER_EXPORTED_EVENT,
        {
            "number_of_exported_events": number_of_exported_events,
            "tracker_store": type(tracker_store).__name__,
            "event_broker": type(event_broker).__name__,
        },
    )


@ensure_telemetry_enabled
def track_interactive_learning_start(
    skip_visualization: bool, save_in_e2e: bool
) -> None:
    """Track when a user starts an interactive learning session.

    Args:
        skip_visualization: Is visualization skipped in this session
        save_in_e2e: Is e2e used in this session
    """
    _track(
        TELEMETRY_INTERACTIVE_LEARNING_STARTED_EVENT,
        {"skip_visualization": skip_visualization, "save_in_e2e": save_in_e2e},
    )


@ensure_telemetry_enabled
def track_server_start(
    input_channels: List["InputChannel"],
    endpoints: Optional["AvailableEndpoints"],
    model_directory: Optional[Text],
    number_of_workers: int,
    is_api_enabled: bool,
) -> None:
    """Tracks when a user starts a rasa server.

    Args:
        input_channels: Used input channels
        endpoints: Endpoint configuration for the server
        model_directory: directory of the running model
        number_of_workers: number of used Sanic workers
        is_api_enabled: whether the rasa API server is enabled
    """
    from rasa.core.utils import AvailableEndpoints

    def project_fingerprint_and_assistant_id_from_model(
        _model_directory: Optional[Text],
    ) -> Tuple[Optional[Text], Optional[Text]]:
        """Gets project fingerprint from an app's loaded model."""
        if not model_directory:
            return None, None

        try:
            model_archive = model.get_local_model(_model_directory)
            metadata = LocalModelStorage.metadata_from_archive(model_archive)

            return metadata.project_fingerprint, metadata.assistant_id
        except Exception:
            return None, None

    if not endpoints:
        endpoints = AvailableEndpoints()

    project, assistant_id = project_fingerprint_and_assistant_id_from_model(
        model_directory
    )

    _track(
        TELEMETRY_SERVER_STARTED_EVENT,
        {
            "input_channels": [i.name() for i in input_channels],
            "api_enabled": is_api_enabled,
            "number_of_workers": number_of_workers,
            "endpoints_nlg": endpoints.nlg.type if endpoints.nlg else None,
            "endpoints_nlu": endpoints.nlu.type if endpoints.nlu else None,
            "endpoints_action_server": endpoints.action.type
            if endpoints.action
            else None,
            "endpoints_model_server": endpoints.model.type if endpoints.model else None,
            "endpoints_tracker_store": endpoints.tracker_store.type
            if endpoints.tracker_store
            else None,
            "endpoints_lock_store": endpoints.lock_store.type
            if endpoints.lock_store
            else None,
            "endpoints_event_broker": endpoints.event_broker.type
            if endpoints.event_broker
            else None,
            "project": project,
            "assistant_id": assistant_id,
        },
    )


@ensure_telemetry_enabled
def track_project_init(path: Text) -> None:
    """Track when a user creates a project using rasa init.

    Args:
        path: Location of the project
    """
    _track(
        TELEMETRY_PROJECT_CREATED_EVENT, {"init_directory": _hash_directory_path(path)}
    )


@ensure_telemetry_enabled
def track_shell_started(model_type: Text, assistant_id: Text) -> None:
    """Track when a user starts a bot using rasa shell.

    Args:
        model_type: Type of the model, core / nlu or rasa.
    """
    _track(
        TELEMETRY_SHELL_STARTED_EVENT,
        {"type": model_type, "assistant_id": assistant_id},
    )


@ensure_telemetry_enabled
def track_inspect_started(channel: Text, assistant_id: Text) -> None:
    """Track when a user starts a bot using rasa inspect.

    Args:
        channel: Type of channel used.
        assistant_id: ID of the assistant being inspected.
    """
    _track(
        TELEMETRY_INSPECT_STARTED_EVENT,
        {"type": channel, "assistant_id": assistant_id},
    )


@ensure_telemetry_enabled
def track_visualization() -> None:
    """Track when a user runs the visualization."""
    _track(TELEMETRY_VISUALIZATION_STARTED_EVENT)


@ensure_telemetry_enabled
def track_core_model_test(num_story_steps: int, e2e: bool, agent: "Agent") -> None:
    """Track when a user tests a core model.

    Args:
        num_story_steps: Number of test stories used for the comparison
        e2e: indicator if tests running in end to end mode
        agent: Agent of the model getting tested
    """
    if agent.processor is None:
        project_fingerprint = ""
    else:
        project_fingerprint = agent.processor.model_metadata.project_fingerprint

    _track(
        TELEMETRY_TEST_CORE_EVENT,
        {
            "project": project_fingerprint,
            "end_to_end": e2e,
            "num_story_steps": num_story_steps,
        },
    )


@ensure_telemetry_enabled
def track_nlu_model_test(test_data: "TrainingData") -> None:
    """Track when a user tests an nlu model.

    Args:
        test_data: Data used for testing
    """
    _track(
        TELEMETRY_TEST_NLU_EVENT,
        {
            "num_intent_examples": len(test_data.intent_examples),
            "num_entity_examples": len(test_data.entity_examples),
            "num_lookup_tables": len(test_data.lookup_tables),
            "num_synonyms": len(test_data.entity_synonyms),
            "num_regexes": len(test_data.regex_features),
        },
    )


@ensure_telemetry_enabled
def track_markers_extraction_initiated(
    strategy: Text, only_extract: bool, seed: bool, count: Optional[int]
) -> None:
    """Track when a user tries to extract success markers.

    Args:
        strategy: The strategy the user is using for tracker selection
        only_extract: Indicates if the user is only extracting markers or also
                      producing stats
        seed: Indicates if the user used a seed for this attempt
        count: (Optional) The number of trackers the user is trying to select.
    """
    _track(
        TELEMETRY_MARKERS_EXTRACTION_INITIATED_EVENT,
        {
            "strategy": strategy,
            "only_extract": only_extract,
            "seed": seed,
            "count": count,
        },
    )


@ensure_telemetry_enabled
def track_markers_extracted(trackers_count: int) -> None:
    """Track when markers have been extracted by a user.

    Args:
        trackers_count: The actual number of trackers processed
    """
    _track(TELEMETRY_MARKERS_EXTRACTED_EVENT, {"trackers_count": trackers_count})


@ensure_telemetry_enabled
def track_markers_stats_computed(trackers_count: int) -> None:
    """Track when stats over markers have been computed by a user.

    Args:
        trackers_count: The actual number of trackers processed
    """
    _track(TELEMETRY_MARKERS_STATS_COMPUTED_EVENT, {"trackers_count": trackers_count})


@ensure_telemetry_enabled
def track_markers_parsed_count(
    marker_count: int, max_depth: int, branching_factor: int
) -> None:
    """Track when markers have been successfully parsed from config.

    Args:
        marker_count: The number of markers found in the config
        max_depth: The maximum depth of any marker in the config
        branching_factor: The maximum number of children of any marker in the config.
    """
    _track(
        TELEMETRY_MARKERS_PARSED_COUNT,
        {
            "marker_count": marker_count,
            "max_depth": max_depth,
            "branching_factor": branching_factor,
        },
    )


def extract_assertion_type_counts(
    input_test_cases: List["TestCase"],
) -> typing.Tuple[bool, Dict[str, Any]]:
    """Extracts the total count of different assertion types from the test cases."""
    from rasa.e2e_test.assertions import AssertionType

    uses_assertions = False

    flow_started_count = 0
    flow_completed_count = 0
    flow_cancelled_count = 0
    pattern_clarification_contains_count = 0
    action_executed_count = 0
    slot_was_set_count = 0
    slot_was_not_set_count = 0
    bot_uttered_count = 0
    generative_response_is_relevant_count = 0
    generative_response_is_grounded_count = 0

    for test_case in input_test_cases:
        for step in test_case.steps:
            assertions = step.assertions if step.assertions else []
            for assertion in assertions:
                if assertion.type == AssertionType.ACTION_EXECUTED.value:
                    action_executed_count += 1
                elif assertion.type == AssertionType.SLOT_WAS_SET.value:
                    slot_was_set_count += 1
                elif assertion.type == AssertionType.SLOT_WAS_NOT_SET.value:
                    slot_was_not_set_count += 1
                elif assertion.type == AssertionType.BOT_UTTERED.value:
                    bot_uttered_count += 1
                elif (
                    assertion.type
                    == AssertionType.GENERATIVE_RESPONSE_IS_RELEVANT.value
                ):
                    generative_response_is_relevant_count += 1
                elif (
                    assertion.type
                    == AssertionType.GENERATIVE_RESPONSE_IS_GROUNDED.value
                ):
                    generative_response_is_grounded_count += 1
                elif assertion.type == AssertionType.FLOW_STARTED.value:
                    flow_started_count += 1
                elif assertion.type == AssertionType.FLOW_COMPLETED.value:
                    flow_completed_count += 1
                elif assertion.type == AssertionType.FLOW_CANCELLED.value:
                    flow_cancelled_count += 1
                elif (
                    assertion.type == AssertionType.PATTERN_CLARIFICATION_CONTAINS.value
                ):
                    pattern_clarification_contains_count += 1

                uses_assertions = True

    result = {
        "flow_started_count": flow_started_count,
        "flow_completed_count": flow_completed_count,
        "flow_cancelled_count": flow_cancelled_count,
        "pattern_clarification_contains_count": pattern_clarification_contains_count,
        "action_executed_count": action_executed_count,
        "slot_was_set_count": slot_was_set_count,
        "slot_was_not_set_count": slot_was_not_set_count,
        "bot_uttered_count": bot_uttered_count,
        "generative_response_is_relevant_count": generative_response_is_relevant_count,
        "generative_response_is_grounded_count": generative_response_is_grounded_count,
    }

    return uses_assertions, result


@ensure_telemetry_enabled
def track_e2e_test_run(
    input_test_cases: List["TestCase"],
    input_fixtures: List["Fixture"],
    input_metadata: List["Metadata"],
) -> None:
    """Track an end-to-end test run."""
    properties = {
        "number_of_test_cases": len(input_test_cases),
        "number_of_fixtures": len(input_fixtures),
        "uses_fixtures": len(input_fixtures) > 0,
        "uses_metadata": len(input_metadata) > 0,
        "number_of_metadata": len(input_metadata),
    }

    uses_assertions, assertion_type_counts = extract_assertion_type_counts(
        input_test_cases
    )

    properties.update({"uses_assertions": uses_assertions})

    if uses_assertions:
        properties.update(assertion_type_counts)

    _track(
        TELEMETRY_E2E_TEST_RUN_STARTED_EVENT,
        properties,
    )


@ensure_telemetry_enabled
def track_response_rephrase(
    rephrase_all: bool,
    custom_prompt_template: Optional[str],
    llm_type: Optional[str],
    llm_model: Optional[str],
    llm_model_group_id: Optional[str],
) -> None:
    """Track when a user rephrases a response."""
    _track(
        TELEMETRY_RESPONSE_REPHRASED_EVENT,
        {
            "rephrase_all": rephrase_all,
            "custom_prompt_template": custom_prompt_template,
            "llm_type": llm_type,
            "llm_model": llm_model,
            "llm_model_group_id": llm_model_group_id,
        },
    )


@ensure_telemetry_enabled
def track_intentless_policy_train() -> None:
    """Track when a user trains a policy."""
    _track(TELEMETRY_INTENTLESS_POLICY_TRAINING_STARTED_EVENT)


@ensure_telemetry_enabled
def track_intentless_policy_train_completed(
    embeddings_type: Optional[str],
    embeddings_model: Optional[str],
    embeddings_model_group_id: Optional[str],
    llm_type: Optional[str],
    llm_model: Optional[str],
    llm_model_group_id: Optional[str],
) -> None:
    """Track when a user trains a policy."""
    _track(
        TELEMETRY_INTENTLESS_POLICY_TRAINING_COMPLETED_EVENT,
        {
            "embeddings_type": embeddings_type,
            "embeddings_model": embeddings_model,
            "embeddings_model_group_id": embeddings_model_group_id,
            "llm_type": llm_type,
            "llm_model": llm_model,
            "llm_model_group_id": llm_model_group_id,
        },
    )


@ensure_telemetry_enabled
def track_intentless_policy_predict(
    embeddings_type: Optional[str],
    embeddings_model: Optional[str],
    embeddings_model_group_id: Optional[str],
    llm_type: Optional[str],
    llm_model: Optional[str],
    llm_model_group_id: Optional[str],
    score: float,
) -> None:
    """Track when a user trains a policy."""
    _track(
        TELEMETRY_INTENTLESS_POLICY_PREDICT_EVENT,
        {
            "embeddings_type": embeddings_type,
            "embeddings_model": embeddings_model,
            "embeddings_model_group_id": embeddings_model_group_id,
            "llm_type": llm_type,
            "llm_model": llm_model,
            "llm_model_group_id": llm_model_group_id,
            "score": score,
        },
    )


@ensure_telemetry_enabled
def identify_endpoint_config_traits(
    endpoints_file: Optional[Text],
    context: Optional[Dict[Text, Any]] = None,
) -> None:
    """Collect traits if enabled.

    Otherwise, sets traits to None.
    """
    traits: Dict[str, Any] = {}

    traits = append_tracing_trait(traits, endpoints_file)
    traits = append_metrics_trait(traits, endpoints_file)
    traits = append_anonymization_trait(traits, endpoints_file)

    _identify(traits, context)


def append_tracing_trait(
    traits: Dict[str, Any], endpoints_file: Optional[str]
) -> Dict[str, Any]:
    """Append the tracing trait to the traits dictionary."""
    import rasa.utils.endpoints
    from rasa.tracing.constants import ENDPOINTS_TRACING_KEY

    tracing_config = rasa.utils.endpoints.read_endpoint_config(
        endpoints_file, ENDPOINTS_TRACING_KEY
    )
    traits[TRACING_BACKEND] = (
        tracing_config.type if tracing_config is not None else None
    )

    return traits


def append_metrics_trait(
    traits: Dict[str, Any], endpoints_file: Optional[str]
) -> Dict[str, Any]:
    """Append the metrics trait to the traits dictionary."""
    import rasa.utils.endpoints
    from rasa.tracing.constants import ENDPOINTS_METRICS_KEY

    metrics_config = rasa.utils.endpoints.read_endpoint_config(
        endpoints_file, ENDPOINTS_METRICS_KEY
    )
    traits[METRICS_BACKEND] = (
        metrics_config.type if metrics_config is not None else None
    )

    return traits


def append_anonymization_trait(
    traits: Dict[str, Any], endpoints_file: Optional[str]
) -> Dict[str, Any]:
    """Append the anonymization trait to the traits dictionary."""
    from rasa.anonymization.anonymisation_rule_yaml_reader import (
        KEY_ANONYMIZATION_RULES,
    )

    anonymization_config = rasa.anonymization.utils.read_endpoint_config(
        endpoints_file, KEY_ANONYMIZATION_RULES
    )

    traits[KEY_ANONYMIZATION_RULES] = (
        rasa.anonymization.utils.extract_anonymization_traits(
            anonymization_config, KEY_ANONYMIZATION_RULES
        )
    )

    return traits


@ensure_telemetry_enabled
def track_enterprise_search_policy_train_started() -> None:
    """Track when a user starts training Enterprise Search policy."""
    _track(TELEMETRY_ENTERPRISE_SEARCH_POLICY_TRAINING_STARTED_EVENT)


@ensure_telemetry_enabled
def track_enterprise_search_policy_train_completed(
    vector_store_type: Optional[str],
    embeddings_type: Optional[str],
    embeddings_model: Optional[str],
    embeddings_model_group_id: Optional[str],
    llm_type: Optional[str],
    llm_model: Optional[str],
    llm_model_group_id: Optional[str],
    citation_enabled: Optional[bool],
) -> None:
    """Track when a user completes training Enterprise Search policy."""
    _track(
        TELEMETRY_ENTERPRISE_SEARCH_POLICY_TRAINING_COMPLETED_EVENT,
        {
            "vector_store_type": vector_store_type,
            "embeddings_type": embeddings_type,
            "embeddings_model": embeddings_model,
            "embeddings_model_group_id": embeddings_model_group_id,
            "llm_type": llm_type,
            "llm_model": llm_model,
            "llm_model_group_id": llm_model_group_id,
            "citation_enabled": citation_enabled,
        },
    )


@ensure_telemetry_enabled
def track_enterprise_search_policy_predict(
    vector_store_type: Optional[str],
    embeddings_type: Optional[str],
    embeddings_model: Optional[str],
    embeddings_model_group_id: Optional[str],
    llm_type: Optional[str],
    llm_model: Optional[str],
    llm_model_group_id: Optional[str],
    citation_enabled: Optional[bool],
) -> None:
    """Track when a user predicts the next action using Enterprise Search policy."""
    _track(
        TELEMETRY_ENTERPRISE_SEARCH_POLICY_PREDICT_EVENT,
        {
            "vector_store_type": vector_store_type,
            "embeddings_type": embeddings_type,
            "embeddings_model": embeddings_model,
            "embeddings_model_group_id": embeddings_model_group_id,
            "llm_type": llm_type,
            "llm_model": llm_model,
            "llm_model_group_id": llm_model_group_id,
            "citation_enabled": citation_enabled,
        },
    )


@ensure_telemetry_enabled
def track_conversation_count_hard_limit(
    conversation_count: int, tracked_month: datetime
) -> None:
    """Track when the number of conversations reaches the hard limit."""
    _track(
        TELEMETRY_CONVERSATION_HARD_LIMIT_REACHED,
        {
            "conversation_count": conversation_count,
            "year": tracked_month.year,
            "month": tracked_month.month,
        },
    )


@ensure_telemetry_enabled
def track_conversation_count_soft_limit(
    conversation_count: int, tracked_month: datetime
) -> None:
    """Track when the number of conversations reaches the soft limit."""
    _track(
        TELEMETRY_CONVERSATION_SOFT_LIMIT_REACHED,
        {
            "conversation_count": conversation_count,
            "year": tracked_month.year,
            "month": tracked_month.month,
        },
    )


@ensure_telemetry_enabled
def track_conversation_count(conversation_count: int, tracked_month: datetime) -> None:
    """Track the number of conversations."""
    _track(
        TELEMETRY_CONVERSATION_COUNT,
        {
            "conversation_count": conversation_count,
            "year": tracked_month.year,
            "month": tracked_month.month,
        },
    )


@ensure_telemetry_enabled
def track_e2e_test_conversion_completed(file_type: str, test_case_count: int) -> None:
    """Track the used input file type for E2E test conversion."""
    _track(
        TELEMETRY_E2E_TEST_CONVERSION_EVENT,
        {
            E2E_TEST_CONVERSION_FILE_TYPE: file_type,
            E2E_TEST_CONVERSION_TEST_CASE_COUNT: test_case_count,
        },
    )


def track_validation_error_log(
    _: structlog.BoundLogger, method_name: str, event_dict: Dict
) -> Dict:
    """Track validation errors."""
    if is_telemetry_enabled() and method_name == "error":
        event_dict_copy = event_dict.copy()

        _track(
            TELEMETRY_VALIDATION_ERROR_LOG_EVENT,
            {
                "log_level": method_name,
                "log_id": event_dict_copy.pop("event"),
                "message": event_dict_copy.pop("event_info", ""),
                **event_dict_copy,
            },
        )
    return event_dict


@ensure_telemetry_enabled
def track_upload_to_studio_failed(response_json: Dict[str, Any]) -> None:
    """Track when the upload to studio fails."""
    _track(
        TELEMETRY_UPLOAD_TO_STUDIO_FAILED_EVENT,
        {"studio_response_json": response_json},
    )
